import numpy as np

from ..constants import int_, float_, complex_, fp_eps
from ..utils.em import dft_spectrum

class SourceTime(object):
    """Base class for the time dependence of a source. Available subclasses:

        - :class:`.GaussianPulse`
    """
    def __init__(self):
        """Construct.
        """

        # NB: tind_beg and tind_end are currently not used but in the future 
        # we could allow a starting and stopping time.
        self.tind_beg = 0
        self.tind_end = None


    def _get_spectrum(self, freqs, tmesh=None):
        """Compute the spectrum of the source in the most general case. 
        This uses a brute-fource DFT which may become slow for many 
        frequencies and time points. For evenly spaced frequencies, we could 
        implement a CZT transform in the future. For an arbitrary selection of 
        frequencies, we could also consider interpolating an FFT.
        
        Parameters
        ----------
        freqs : array_like
            Array of frequencies to sample the spectrum at.
        tmesh : array_like
            This function does a numerical computation of the spectrum and so 
            requires a mesh on which the time-dependence is computed and 
            transformed. It can be overriden by methods that use an analytic 
            result, in which case ``tmesh is not needed.

        Returns
        -------
        spectrum : array_like
            Array of same size as ``freqs`` giving the complex-valued spectrum 
            of the source.
        """

        if tmesh is None or len(tmesh)<=1:
            return None

        dt = tmesh[1] - tmesh[0]
        tdep = self._get_time(tmesh)
        return dft_spectrum(tdep, dt, freqs)

class GaussianPulse(SourceTime):
    """ Source with a Gaussian-envelope time dependence.
    """

    def __init__(self, frequency, fwidth, offset=5., phase=0):
        """ Construct.
        
        Parameters
        ----------
        frequency : float
            (Hertz) Carrier frequency.
        fwidth : float
            (Hertz) Frequency bandwidth.
        offset : float, optional
            Offset of the peak of the Gaussian envelope from the starting time 
            of the simulation, in units of ``twidth = 1/2/pi/fwidth``. 
            The peak of the source amplitude is attained at time 
            ``t = offset/2/pi/fwidth``. 
        """
        super().__init__()
        self.frequency = frequency
        self.fwidth = fwidth
        self.twidth = 1 / 2 / np.pi / fwidth
        self.offset = offset
        # User-defined global phase
        self.phase = phase

    def __repr__(self):
        rep = "GaussianPulse(\n"
        rep += "    frequency  = %1.2e,\n"%self.frequency
        rep += "    fwidth     = %1.2e,\n"%self.fwidth
        rep += "    offset     = %1.2f)"%self.offset

        return rep

    def _get_time(self, tmesh, extra_phase=0):
        """ Compute the time dependence of the source over ``tmesh``, using a 
        first-derivative of a Gaussian for the envelope.
        """

        # Global phase plus extra phase
        phi = self.phase + extra_phase
        tt0 = tmesh - self.offset * self.twidth
        omega = 2*np.pi*self.frequency
        G = (1 + tmesh / self.twidth**2 / omega) * \
                np.exp(1j*phi - 1j*omega*tmesh - tt0**2/2/self.twidth**2)
                
        G = np.real(G)

        # Starting and stopping time, currently unused.
        Gnz = np.nonzero(np.abs(G) > 1e-5)[0]
        if Gnz.size > 0:
            self.tind_beg = Gnz[0]
            self.tind_end = Gnz[-1]
        else:
            self.tind_beg = 0
            self.tind_end = 0

        return G

    # def _get_spectrum(self, freqs, tmesh=None):
    #     """For a Gaussian pulse, it should be possible to compute the spectrum 
    #     analytically. However, I keep getting a weird (not just a prefactor) 
    #     mismatch between the formulat and the numerical result. The difference 
    #     depends both on fwidht and offset of the source.
        
    #     Parameters
    #     ----------
    #     freqs : array_like
    #         Array of frequencies to sample the spectrum at.
    #     """

    #     fr  = np.array(freqs) / self.fwidth # normalized frequences
    #     fr0 = self.frequency / self.fwidth  # normalized center frequency
    #     return self.twidth * np.exp(1j*(fr-fr0)*self.offset - 0.5*(fr-fr0)**2)


import numpy as np

from ..utils import inside_box, inside_box_inds, cs2span, listify, list2str
from ..utils import log_and_raise
from ..utils.check import check_3D_lists
from ..constants import int_, float_, fp_eps, xyz_list, xyz_dict, C_0

class Monitor(object):

    def __init__(self, center, size, field, name=None):
        """Base class for structures. Available subclasses:

        - :class:`.TimeMonitor`
        - :class:`.FreqMonitor`
        - :class:`.ModeMonitor`
        """

        check_3D_lists(center=listify(center), size=listify(size))
        self.center = np.array(center)
        self.size = np.array(size)
        self.span = cs2span(self.center, self.size)
        self.field = []
        for f in listify(field):
            if f.lower() in ['e', 'h']:
                self.field.append(f.lower())
            else:
                log_and_raise(
                    f"Unrecognized field {f}. Valid values are 'E' and 'H'.",
                    ValueError
                )

        self.name = None if name is None else str(name)

    def _inside(self, mesh):
        """ Get a mask equal to one if a point is inside the monitor region, 
        and zero if outside.
        
        Parameters
        ----------
        mesh : 3-tuple
            Defines the x, y, and z mesh. 
        """
        return inside_box(self.span, mesh, include_zero_size=True)

    def _inside_inds(self, mesh):
        """ Get indexes of the points at which the fields are to be recored.
        
        Parameters
        ----------
        mesh : 3-tuple
            Defines the x, y, and z mesh. 
        
        Returns
        -------
        Pinds : np.ndarray
            An array of shape (Np, 3), where Np is the total number of mesh 
            points in the monitor region.
        """
        inds = inside_box_inds(self.span, mesh, include_zero_size=True)
        indsm = np.meshgrid(np.arange(inds[0][0], inds[0][1]), 
                            np.arange(inds[1][0], inds[1][1]),
                            np.arange(inds[2][0], inds[2][1]),
                            indexing='ij')
        if indsm[0].size==0: 
            return np.zeros((0, 3), dtype=int_)

        return np.stack([inds.ravel() for inds in indsm], axis=1).astype(int_)

class TimeMonitor(Monitor):
    """Monitor recording the time-domain fields within a 3D region.
    """

    def __init__(self, center, size, t_start=0, t_stop=None, 
                    field=['E', 'H'], name=None):
        """ Construct.
        
        Parameters
        ----------
        center : array_like
            (micron) x, y, and z position of the center of the Monitor.
        size : array_like
            (micron) Size in x, y, and z.
        t_start : float, optional
            (second) Starting time of field recording.
        t_stop : float, optional
            (second) Stopping time of field recording. If ``None``, record 
            until the end of the simulation.
        field : list, optional
            List of fields to be recorded. Valid entries are 
            ``'E'`` and ``'H'``.
        name : str, optional
            Custom name of the monitor.

        Note
        ----
        Time monitors can result in very large amounts of data if defined over 
        a large spatial region. Recommended usage is either recording the full 
        time evolution of a single point in space, or using ``t_start`` and 
        ``t_stop`` to record just a few time steps of a larger region. 
        """
        super().__init__(center, size, field)
        self.t_start = t_start
        self.t_stop = t_stop

    def __repr__(self):
        rep = "Tidy3D TimeMonitor: {\n"
        rep += "name     = %s\n"%self.name
        rep += "center   = %s\n" % list2str(self.center, "%1.4f")
        rep += "size     = %s\n" % list2str(self.size, "%1.4f")
        rep += "t_start  = %1.2e,\n"%self.t_start
        if self.t_stop is None:
            rep += "t_stop   = None\n"
        else:
            rep += "t_stop   = %1.2e,\n"%self.t_stop

        rep += "Store E:  %s\n"%('e' in self.field)
        rep += "Store H:  %s\n"%('h' in self.field)
        rep += "}\n"

        return rep

class FreqMonitor(Monitor):
    """Monitor recording a discrete Fourier transform of the fields within a 
    3D region, for a given list of frequencies.
    """
    
    def __init__(self, center, size, freqs, field=['E', 'H'], name=None):
        """ Construct.

        Parameters
        ----------
        center : array_like
            (micron) x, y, and z position of the center of the Monitor.
        size : array_like
            (micron) Size in x, y, and z.
        freqs : float or array_like
            Frequencies at which the fields are sampled.
        field : list, optional
            List of fields to be recorded. Valid entries are 
            ``'E'`` and ``'H'``.
        name : str, optional
            Custom name of the monitor.
        """

        super().__init__(center, size, field, name)
        self.freqs = listify(freqs)
        self.lambdas = C_0 / np.array(freqs)

    def __repr__(self):
        rep = "Tidy3D FreqMonitor: {\n"
        rep += "name     = %s\n"%self.name
        rep += "center   = %s\n" % list2str(self.center, "%1.4f")
        rep += "size     = %s\n" % list2str(self.size, "%1.4f")
        rep += "freqs    = %s\n" % list2str(self.freqs, "%1.2e")

        rep += "Store E:  %s\n"%('e' in self.field)
        rep += "Store H:  %s\n"%('h' in self.field)
        rep += "}\n"

        return rep

class ModeMonitor(FreqMonitor):
    """ :class:`.FreqMonitor` subclass defining a 2D plane in which the recorded 
    frequency-domain fields can be decomposed into propagating eigenmodes.
    """
    
    def __init__(self, center, size, freqs, name=None):
        """ Construct.

        Parameters
        ----------
        center : array_like
            (micron) 3D vector defining the center of the 2D plane.
        size : array_like
            (micron) 3D vector defining the size of the 2D plane. Exactly one 
            of the values must be ``0``, defining the normal direction.
        freqs : float or list of float
            Frequencies at which the fields are sampled.
        name : str, optional
            Custom name of the monitor.
        """

        # Get normal direction from size
        self.norm_ind = np.nonzero(size < fp_eps)[0]
        if self.norm_ind.size !=1:
            log_and_raise(
                "Exactly one element of ModeMonitor 'size' must be zero.",
                ValueError
            )
            
        self.norm_ind = int(self.norm_ind)
        self.normal = xyz_list[self.norm_ind]

        super().__init__(center, size, freqs, field=['E', 'H'], name=name)

    def __repr__(self):
        rep = "Tidy3D ModeMonitor: {\n"
        rep += "name     = %s\n"%self.name
        rep += "normal   = %s\n"%self.normal
        rep += "center   = %s\n" % list2str(self.center, "%1.4f")
        rep += "size     = %s\n" % list2str(self.size, "%1.4f")
        rep += "freqs    = %s\n" % list2str(self.freqs, "%1.2e")
        rep += "}\n"

        return rep


import numpy as np
import scipy.sparse as sp

from ..constants import EPSILON_0, ETA_0

def compute_derivative_matrices(omega, shape, npml, mesh_step):
    """Construct sparse matrices applying finite-difference derivatives in 2D.

    Parameters
    ----------
    omega : float
        (Hertz) Angular frequency.
    shape : tuple of int
        Size (Nx, Ny) of the 2D domain.
    npml : tuple of int
        Number of PML layers in x and y.
    mesh_step : tuple of float
        (micron) Mesh step in x and y.
    
    Returns
    -------
    scipy.sparse matrix
        Matrices implementing the forward and backward derivatives along x and 
        y, respectively.
    """

    dx, dy = mesh_step

    # Construct derivate matrices without PML.
    Dxf = createDws('x', 'f', shape, dx)
    Dxb = createDws('x', 'b', shape, dx)
    Dyf = createDws('y', 'f', shape, dy)
    Dyb = createDws('y', 'b', shape, dy)

    # Make the S-matrices for PML.
    (Sxf, Sxb, Syf, Syb) = create_S_matrices(omega, shape, npml, mesh_step)

    # Apply PML to derivative matrices.
    Dxf = Sxf.dot(Dxf)
    Dxb = Sxb.dot(Dxb)
    Dyf = Syf.dot(Dyf)
    Dyb = Syb.dot(Dyb)

    return Dxf, Dxb, Dyf, Dyb

def createDws(component, direction, shape, dL):
    """Derivative matrices without PML.
    
    Parameters
    ----------
    component : str
        One of {'x', 'y'}.
    direction : str
        One of {'f', 'b'}.
    shape : tuple of int
        Size (Nx, Ny) of the 2D domain.
    dL : float
        Mesh step in the specified direction.
    """

    Nx, Ny = shape

    # special case, a 1D problem
    if component == 'x' and Nx == 1:
        return sp.eye(Ny)
    if component == 'y' and Ny == 1:
        return sp.eye(Nx)

    # select a `make_D` function based on the component and direction
    component_dir = component + direction
    if component_dir == 'xf':
        return make_Dxf(dL, shape)
    elif component_dir == 'xb':
        return make_Dxb(dL, shape)
    elif component_dir == 'yf':
        return make_Dyf(dL, shape)
    elif component_dir == 'yb':
        return make_Dyb(dL, shape)
    else:
        raise ValueError("Component %s and/or direction %s not recognized."%(
                            component, direction))

def make_Dxf(dL, shape):
    """ Forward derivative in x. """
    Nx, Ny = shape
    Dxf = sp.diags([-1, 1, 1], [0, 1, -Nx+1], shape=(Nx, Nx))
    Dxf = 1 / dL * sp.kron(Dxf, sp.eye(Ny))
    return Dxf

def make_Dxb(dL, shape):
    """ Backward derivative in x. """
    Nx, Ny = shape
    Dxb = sp.diags([1, -1, -1], [0, -1, Nx-1], shape=(Nx, Nx))
    Dxb = 1 / dL * sp.kron(Dxb, sp.eye(Ny))
    return Dxb

def make_Dyf(dL, shape):
    """ Forward derivative in y. """
    Nx, Ny = shape
    Dyf = sp.diags([-1, 1, 1], [0, 1, -Ny+1], shape=(Ny, Ny))
    Dyf = 1 / dL * sp.kron(sp.eye(Nx), Dyf)
    return Dyf

def make_Dyb(dL, shape):
    """ Backward derivative in y. """
    Nx, Ny = shape
    Dyb = sp.diags([1, -1, -1], [0, -1, Ny-1], shape=(Ny, Ny))
    Dyb = 1 / dL * sp.kron(sp.eye(Nx), Dyb)
    return Dyb

def create_S_matrices(omega, shape, npml, mesh_step):
    """Makes the 'S-matrices'. When dotted with derivative matrices, they add 
    PML. """

    # strip out some information needed
    Nx, Ny = shape
    dx, dy = mesh_step
    N = Nx * Ny
    x_range = [0, float(dx * Nx)]
    y_range = [0, float(dy * Ny)]
    Nx_pml, Ny_pml = npml    

    # Create the sfactor in each direction and for 'f' and 'b'
    s_vector_x_f = create_sfactor('f', omega, dx, Nx, Nx_pml)
    s_vector_x_b = create_sfactor('b', omega, dx, Nx, Nx_pml)
    s_vector_y_f = create_sfactor('f', omega, dy, Ny, Ny_pml)
    s_vector_y_b = create_sfactor('b', omega, dy, Ny, Ny_pml)

    # Fill the 2D space with layers of appropriate s-factors
    Sx_f_2D = np.zeros(shape, dtype=np.complex128)
    Sx_b_2D = np.zeros(shape, dtype=np.complex128)
    Sy_f_2D = np.zeros(shape, dtype=np.complex128)
    Sy_b_2D = np.zeros(shape, dtype=np.complex128)

    # Insert the cross sections into the S-grids (could be done more elegantly)
    for i in range(0, Ny):
        Sx_f_2D[:, i] = 1 / s_vector_x_f
        Sx_b_2D[:, i] = 1 / s_vector_x_b
    for i in range(0, Nx):
        Sy_f_2D[i, :] = 1 / s_vector_y_f
        Sy_b_2D[i, :] = 1 / s_vector_y_b

    # Reshape the 2D s-factors into a 1D s-vecay
    Sx_f_vec = Sx_f_2D.flatten()
    Sx_b_vec = Sx_b_2D.flatten()
    Sy_f_vec = Sy_f_2D.flatten()
    Sy_b_vec = Sy_b_2D.flatten()

    # Construct the 1D total s-vector into a diagonal matrix
    Sx_f = sp.spdiags(Sx_f_vec, 0, N, N)
    Sx_b = sp.spdiags(Sx_b_vec, 0, N, N)
    Sy_f = sp.spdiags(Sy_f_vec, 0, N, N)
    Sy_b = sp.spdiags(Sy_b_vec, 0, N, N)

    return Sx_f, Sx_b, Sy_f, Sy_b

def create_sfactor(direction, omega, dL, N, N_pml):
    """ Creates the S-factor cross section needed in the S-matrices """

    # For no PNL, this should just be identity matrix.
    if N_pml == 0:
        return np.ones(N, dtype=np.complex128)

    # Otherwise, get different profiles for forward and reverse derivatives.
    if direction == 'f':
        return create_sfactor_f(omega, dL, N, N_pml)
    elif direction == 'b':
        return create_sfactor_b(omega, dL, N, N_pml)
    else:
        raise ValueError("Direction value {} not recognized".format(direction))

def create_sfactor_f(omega, dL, N, N_pml):
    """ S-factor profile for forward derivative matrix """
    sfactor_array = np.ones(N, dtype=np.complex128)
    for i in range(N):
        if i <= N_pml:
            sfactor_array[i] = s_value(dL, (N_pml - i + 0.5)/N_pml, omega)
        elif i > N - N_pml:
            sfactor_array[i] = s_value(dL, (i - (N - N_pml) - 0.5)/N_pml, omega)
    return sfactor_array

def create_sfactor_b(omega, dL, N, N_pml):
    """ S-factor profile for backward derivative matrix """
    sfactor_array = np.ones(N, dtype=np.complex128)
    for i in range(N):
        if i <= N_pml:
            sfactor_array[i] = s_value(dL, (N_pml - i + 1)/N_pml, omega)
        elif i > N - N_pml:
            sfactor_array[i] = s_value(dL, (i - (N - N_pml) - 1)/N_pml, omega)
    return sfactor_array

def sig_w(dL, step, sorder=3):
    """ Fictional conductivity, note that these values might need tuning """
    sig_max = 0.8 * (sorder + 1) / (ETA_0 * dL)
    return sig_max * step**sorder

def s_value(dL, step, omega):
    """ S-value to use in the S-matrices """
    # print(step)
    return 1 - 1j * sig_w(dL, step) / (omega * EPSILON_0)



# def compute_derivative_matrices(omega, shape, npml, mesh_step, bloch_x=0.0, bloch_y=0.0):
#     """ Returns sparse derivative matrices.  Currently works for 2D and 1D 
#             omega: angular frequency (rad/sec)
#             shape: shape of the FDFD grid
#             npml: list of number of PML cells in x and y.
#             dL: spatial grid size (m)
#             block_x: bloch phase (phase across periodic boundary) in x
#             block_y: bloch phase (phase across periodic boundary) in y
#     """

#     dL = mesh_step[0]

#     # Construct derivate matrices without PML
#     Dxf = createDws('x', 'f', shape, dL, bloch_x=bloch_x, bloch_y=bloch_y)
#     Dxb = createDws('x', 'b', shape, dL, bloch_x=bloch_x, bloch_y=bloch_y)
#     Dyf = createDws('y', 'f', shape, dL, bloch_x=bloch_x, bloch_y=bloch_y)
#     Dyb = createDws('y', 'b', shape, dL, bloch_x=bloch_x, bloch_y=bloch_y)

#     # make the S-matrices for PML
#     (Sxf, Sxb, Syf, Syb) = create_S_matrices(omega, shape, npml, dL)

#     # apply PML to derivative matrices
#     Dxf = Sxf.dot(Dxf)
#     Dxb = Sxb.dot(Dxb)
#     Dyf = Syf.dot(Dyf)
#     Dyb = Syb.dot(Dyb)

#     return Dxf, Dxb, Dyf, Dyb

# def createDws(component, dir, shape, dL, bloch_x=0.0, bloch_y=0.0):
#     """ creates the derivative matrices
#             component: one of 'x' or 'y' for derivative in x or y direction
#             dir: one of 'f' or 'b', whether to take forward or backward finite difference
#             shape: shape of the FDFD grid
#             dL: spatial grid size (m)
#             block_x: bloch phase (phase across periodic boundary) in x
#             block_y: bloch phase (phase across periodic boundary) in y
#     """

#     Nx, Ny = shape    

#     # special case, a 1D problem
#     if component == 'x' and Nx == 1:
#         return sp.eye(Ny)
#     if component is 'y' and Ny == 1:
#         return sp.eye(Nx)

#     # select a `make_D` function based on the component and direction
#     component_dir = component + dir
#     if component_dir == 'xf':
#         return make_Dxf(dL, shape, bloch_x=bloch_x)
#     elif component_dir == 'xb':
#         return make_Dxb(dL, shape, bloch_x=bloch_x)
#     elif component_dir == 'yf':
#         return make_Dyf(dL, shape, bloch_y=bloch_y)
#     elif component_dir == 'yb':
#         return make_Dyb(dL, shape, bloch_y=bloch_y)
#     else:
#         raise ValueError("component and direction {} and {} not recognized".format(component, dir))

# def make_Dxf(dL, shape, bloch_x=0.0):
#     """ Forward derivative in x """
#     Nx, Ny = shape
#     phasor_x = np.exp(1j * bloch_x)
#     Dxf = sp.diags([-1, 1, phasor_x], [0, 1, -Nx+1], shape=(Nx, Nx), dtype=np.complex128)
#     Dxf = 1 / dL * sp.kron(Dxf, sp.eye(Ny))
#     return Dxf

# def make_Dxb(dL, shape, bloch_x=0.0):
#     """ Backward derivative in x """
#     Nx, Ny = shape
#     phasor_x = np.exp(1j * bloch_x)
#     Dxb = sp.diags([1, -1, -np.conj(phasor_x)], [0, -1, Nx-1], shape=(Nx, Nx), dtype=np.complex128)
#     Dxb = 1 / dL * sp.kron(Dxb, sp.eye(Ny))
#     return Dxb

# def make_Dyf(dL, shape, bloch_y=0.0):
#     """ Forward derivative in y """
#     Nx, Ny = shape
#     phasor_y = np.exp(1j * bloch_y)
#     Dyf = sp.diags([-1, 1, phasor_y], [0, 1, -Ny+1], shape=(Ny, Ny))
#     Dyf = 1 / dL * sp.kron(sp.eye(Nx), Dyf)
#     return Dyf

# def make_Dyb(dL, shape, bloch_y=0.0):
#     """ Backward derivative in y """
#     Nx, Ny = shape
#     phasor_y = np.exp(1j * bloch_y)
#     Dyb = sp.diags([1, -1, -np.conj(phasor_y)], [0, -1, Ny-1], shape=(Ny, Ny))
#     Dyb = 1 / dL * sp.kron(sp.eye(Nx), Dyb)
#     return Dyb


# """ PML Functions """

# def create_S_matrices(omega, shape, npml, dL):
#     """ Makes the 'S-matrices'.  When dotted with derivative matrices, they add PML """

#     # strip out some information needed
#     Nx, Ny = shape
#     N = Nx * Ny
#     x_range = [0, float(dL * Nx)]
#     y_range = [0, float(dL * Ny)]
#     Nx_pml, Ny_pml = npml    

#     # Create the sfactor in each direction and for 'f' and 'b'
#     s_vector_x_f = create_sfactor('f', omega, dL, Nx, Nx_pml)
#     s_vector_x_b = create_sfactor('b', omega, dL, Nx, Nx_pml)
#     s_vector_y_f = create_sfactor('f', omega, dL, Ny, Ny_pml)
#     s_vector_y_b = create_sfactor('b', omega, dL, Ny, Ny_pml)

#     # Fill the 2D space with layers of appropriate s-factors
#     Sx_f_2D = np.zeros(shape, dtype=np.complex128)
#     Sx_b_2D = np.zeros(shape, dtype=np.complex128)
#     Sy_f_2D = np.zeros(shape, dtype=np.complex128)
#     Sy_b_2D = np.zeros(shape, dtype=np.complex128)

#     # insert the cross sections into the S-grids (could be done more elegantly)
#     for i in range(0, Ny):
#         Sx_f_2D[:, i] = 1 / s_vector_x_f
#         Sx_b_2D[:, i] = 1 / s_vector_x_b
#     for i in range(0, Nx):
#         Sy_f_2D[i, :] = 1 / s_vector_y_f
#         Sy_b_2D[i, :] = 1 / s_vector_y_b

#     # Reshape the 2D s-factors into a 1D s-vecay
#     Sx_f_vec = Sx_f_2D.flatten()
#     Sx_b_vec = Sx_b_2D.flatten()
#     Sy_f_vec = Sy_f_2D.flatten()
#     Sy_b_vec = Sy_b_2D.flatten()

#     # Construct the 1D total s-vecay into a diagonal matrix
#     Sx_f = sp.spdiags(Sx_f_vec, 0, N, N)
#     Sx_b = sp.spdiags(Sx_b_vec, 0, N, N)
#     Sy_f = sp.spdiags(Sy_f_vec, 0, N, N)
#     Sy_b = sp.spdiags(Sy_b_vec, 0, N, N)

#     return Sx_f, Sx_b, Sy_f, Sy_b

# def create_sfactor(dir, omega, dL, N, N_pml):
#     """ creates the S-factor cross section needed in the S-matrices """

#     #  for no PNL, this should just be zero
#     if N_pml == 0:
#         return np.ones(N, dtype=np.complex128)

#     # otherwise, get different profiles for forward and reverse derivative matrices
#     dw = N_pml * dL
#     if dir == 'f':
#         return create_sfactor_f(omega, dL, N, N_pml, dw)
#     elif dir == 'b':
#         return create_sfactor_b(omega, dL, N, N_pml, dw)
#     else:
#         raise ValueError("Dir value {} not recognized".format(dir))

# def create_sfactor_f(omega, dL, N, N_pml, dw):
#     """ S-factor profile for forward derivative matrix """
#     sfactor_array = np.ones(N, dtype=np.complex128)
#     for i in range(N):
#         if i <= N_pml:
#             sfactor_array[i] = s_value(dL * (N_pml - i + 0.5), dw, omega)
#         elif i > N - N_pml:
#             sfactor_array[i] = s_value(dL * (i - (N - N_pml) - 0.5), dw, omega)
#     return sfactor_array

# def create_sfactor_b(omega, dL, N, N_pml, dw):
#     """ S-factor profile for backward derivative matrix """
#     sfactor_array = np.ones(N, dtype=np.complex128)
#     for i in range(N):
#         if i <= N_pml:
#             sfactor_array[i] = s_value(dL * (N_pml - i + 1), dw, omega)
#         elif i > N - N_pml:
#             sfactor_array[i] = s_value(dL * (i - (N - N_pml) - 1), dw, omega)
#     return sfactor_array

# def sig_w(l, dw, m=3, lnR=-30):
#     """ Fictional conductivity, note that these values might need tuning """
#     sig_max = -(m + 1) * lnR / (2 * ETA_0 * dw) 
#     return sig_max * (l / dw)**m 

# def s_value(l, dw, omega):
#     """ S-value to use in the S-matrices """
#     return 1 - 1j * sig_w(l, dw) / (omega * EPSILON_0)


import math
import numpy as np
from .constants import C_0, HBAR

class DispersionModel(object):
    """
    Base class for a model of material dispersion.
    All dispersion models are required to fit into the pole-residue model,
    in which
    ``epsilon(omega) = epsilon_infinity +
                       sum_i [      c[i] /(j omega -      a[i])  +
                               conj(c[i])/(j omega - conj(a[i])) ]``
    poles are stored internally in units of rad/s.
    """

    def __init__(self, eps_inf = 1.0, poles = None):
        """ Initialize a raw DispersionModel.
        eps_inf: The relative permittivity at infinite frequency, usually 1.
        poles: A list of the (a, c) coefficients, specified here in rad/s.
        """
        self._eps_inf = 1
        # A list of (a, c) coefficients for the underlying pole-residue model.
        # The units of (a, c) are in angular frequency (rad/s).
        if poles is None:
            self._poles = []
        else:
            self._poles = [(pole[0], pole[1]) for pole in poles]

    def epsilon(self, freq = None):
        """ Evaluate the permittivity at the given frequency (Hz).
        """
        w = 2*np.pi*freq
        eps = self._eps_inf
        for (a,c) in self._poles:
            eps += c/(1j*w-a) + c.conjugate()/(1j*w-a.conjugate())
        return eps

class Sellmeier(DispersionModel):
    def __init__(self, coeffs, name=None):
        """ Define a material with Sellmeier dispersion
        
        Parameters
        ----------
        coeffs : list, of the form [(B0, C0), (B1, C1), ...]
             The dispersion formula is:
             n**2 - 1 = B0*lambda**2 / (lambda**2 - C0) + ...
             where lambda is in microns.
        """
        self._eps_inf = 1
        self._poles = []
        for BC in coeffs:
            # C enters with units of microns**2, B is unitless
            # If we let c = alpha*1j, a = beta*1j, then
            # B = 2*alpha/beta, C = (2*pi*c0/beta)**2
            beta = 2*np.pi*C_0 / np.sqrt(BC[1]) # This has units of rad/s
            alpha = -0.5*beta*BC[0]
            a = 1j*beta
            c = 1j*alpha
            self._poles.append((a, c))

class Lorentz(DispersionModel):
    def __init__(self, eps_inf, coeffs, name=None):
        """ Define a material with Lorentz dispersion
        
        Parameters
        ----------
        coeffs : list, of the form [(Delta_eps_0, f_0, delta_0), ...]
             The dispersion formula is:
             eps(f) = eps_infty + sum_p Delta_eps_p f_p**2 /
                                        (f_p**2 + 2*j*f*delta_p - f**2)
             where f, f_p, delta_p are in Hz.
        """
        self._eps_inf = eps_inf
        self._poles = []
        for c in coeffs:
            w = 2*math.pi*c[1]
            d = 2*math.pi*c[2]
            if d > w:
                r = 1j*np.sqrt(d*d-w*w)
            else:
                r = np.sqrt(w*w - d*d)
            a = -d - 1j*r
            c = 0.5j*c[0]*w*w / r
            self._poles.append((a, c))

class Debye(DispersionModel):
    def __init__(self, eps_inf, coeffs, name=None):
        """ Define a material with Debye dispersion
        
        Parameters
        ----------
        coeffs : list, of the form [(Delta_eps_0, tau_0), ...]
             The dispersion formula is:
             eps(f) = eps_infty + sum_p Delta_eps_p / (1 + j*f*tau_p)
             where f is in Hz, and tau_p is in s.
        """
        self._eps_inf = eps_inf
        self._poles = []
        for c in coeffs:
            a = -2*math.pi / c[1]
            c = -0.5*c[0]*a
            self._poles.append((a, c))

def fit(x, y, x_units='um', y_units='nk', npoles=3):
    """ Fits numerical material dispersion data to a model.
    x : sequence of floats
        Values for x-axis (wavelength, energy, frequency, etc.)
    y : sequence of floats
        Values for y-axis (permittivity, refractive index, etc.)
    x_units : string
        The units of the values in x. Valid units are:
        'um': microns
        'eV': electron-volts
        '1/cm': inverse centimeters
        '1/um': inverse microns
        'Hz': Hertz
    y_units : string
        The units of the values in y. Valid units are:
        'nk': refractive index, real part is n, complex part is k.
              Lossy materials have a positive complex part.
        'eps': permittivity. Lossy materials have a positive complex part.
    npoles : int
        The number of poles in the fitted model. Higher numbers produce
        more accurate models, but require more simulation time. Typically,
        6 poles per decade of x values are used.
    """
    raise NotImplementedError("Fitting is not yet implemented.")

    


import numpy as np
import subprocess

from .constants import fp_eps, EPSILON_0


def listify(obj):
    # Make a list if not a list
    if isinstance(obj, list):
        return obj
    elif isinstance(obj, tuple):
        return list(obj)
    elif isinstance(obj, np.ndarray):
        return obj.tolist()
    else:
        return [obj]


def list2str(li, elem_format):
    # Make a string that looks like the list ``li`` using %-specifying string 
    # ``elem_format`` for each element
    return "[" + " ".join([elem_format % elem for elem in li]) + "]"


def subprocess_cmd(command):
    """Execute a (multi-line) shell command.

    Parameters
    ----------
    command : str
        Semicolon separated lines.
    """
    comm_lines = command.split(";")
    for line in comm_lines:
        comm_list = list(line.split())
        process = subprocess.run(
            comm_list, stdout=None, check=True, stdin=subprocess.DEVNULL
        )


def inside_box_inds(span, mesh, include_zero_size=False):
    """Elementwise indicator function showing which points in a Box ``span``
    are inside a grid defined by ``mesh``.

    Parameters
    ----------
    span : np.ndarray of shape (3, 2)
        Defines (xmin, xmax), (ymin, ymax), (zmin, zmax) of the box.
    mesh : tuple
        3-tuple defining the xgrid, ygrid and zgrid.
    include_zero_size : bool, optional
        If True, a span of zero (max - min = 0) in any direction will be taken
        as one pixel. If False, the mask will be all empty.

    Returns
    -------
    indsx, indsy, indsz: tuple
        Tuples defining the (start, stop) index of the mesh inside the box.
    """

    # Check if min and max values are in order
    for dim in range(3):
        if span[dim, 1] < span[dim, 0]:
            raise ValueError(
                "Incorrect object span (max value smaller than " "min value)."
            )

    # Initialize indexes for no part of the mesh inside the box
    indsx, indsy, indsz = ((0, 0), (0, 0), (0, 0))

    # Do some checks that should return no mesh inside the box
    for dim in range(3):
        # Check if any of the mesh arrays has zero lengt
        if mesh[dim].size == 0:
            return indsx, indsy, indsz
        # Check if box is completely outside mesh
        if (span[dim, 0] > mesh[dim][-1]) or (span[dim, 1] < mesh[dim][0]):
            return indsx, indsy, indsz
        # Check if there's a strictly zero span and it was not requested.
        if (span[dim, 1] - span[dim, 0] == 0) and (not include_zero_size):
            return indsx, indsy, indsz

    indx = np.nonzero((span[0, 0] < mesh[0]) * (mesh[0] < span[0, 1]))[0]
    indy = np.nonzero((span[1, 0] < mesh[1]) * (mesh[1] < span[1, 1]))[0]
    indz = np.nonzero((span[2, 0] < mesh[2]) * (mesh[2] < span[2, 1]))[0]

    # If there's still a zero span and it was not requested, we're done.
    if (indx.size == 0 or indy.size == 0 or indz.size == 0) and (
        not include_zero_size
    ):
        return indsx, indsy, indsz

    # Otherwise, we build the mask. For zero span, we just take the first
    # element of the mesh that is larger than the requested min value.
    if indx.size == 0:
        indx1 = np.nonzero(span[0, 0] <= mesh[0])[0][0]
        indsx = (indx1, indx1 + 1)
    else:
        indsx = (indx[0], indx[-1] + 1)
    if indy.size == 0:
        indy1 = np.nonzero(span[1, 0] <= mesh[1])[0][0]
        indsy = (indy1, indy1 + 1)
    else:
        indsy = (indy[0], indy[-1] + 1)
    if indz.size == 0:
        indz1 = np.nonzero(span[2, 0] <= mesh[2])[0][0]
        indsz = (indz1, indz1 + 1)
    else:
        indsz = (indz[0], indz[-1] + 1)

    return (indsx, indsy, indsz)


def inside_box(span, mesh, include_zero_size=False):
    """Elementwise indicator function showing which points in the ``span`` are
    inside a grid defined by ``mesh``.

    Parameters
    ----------
    span : np.ndarray of shape (3, 2)
        Defines (xmin, xmax), (ymin, ymax), (zmin, zmax) of the box.
    mesh : tuple
        3-tuple defining the xgrid, ygrid and zgrid.
    include_zero_size : bool, optional
        If True, a span of zero (max - min = 0) in any direction will be taken
        as one pixel. If False, the mask will be all empty.

    Returns
    -------
    mask : np.ndarray
        A 3D array of shape (mesh[0].size, mesh[1].size, mesh[2].size)
        that is 1 inside the box and 0 outside.
    """

    # Initialize empty mask
    mask = np.zeros((mesh[0].size, mesh[1].size, mesh[2].size))

    indsx, indsy, indsz = inside_box_inds(span, mesh, include_zero_size)
    mask[indsx[0] : indsx[1], indsy[0] : indsy[1], indsz[0] : indsz[1]] = 1.0

    return mask


def intersect_box(span1, span2):
    """Return a span of a box that is the intersection between two spans."""
    span = np.zeros((3, 2))
    for d in range(3):
        span[d, 0] = max(span1[d, 0], span2[d, 0])
        span[d, 1] = min(span1[d, 1], span2[d, 1])

    return span


def object_name(obj_list, obj, prefix=""):
    """Return a name for the object that is unique within the names in a
    given list. The optional prefix is to be used if object.name is None.
    """
    if obj.name is not None:
        prefix = obj.name

    name_list = [obj.name for obj in obj_list]
    count = 1
    name = prefix
    if obj.name is None:
        name += "_0"

    while name in name_list:
        name = prefix + "_" + str(count)
        count += 1

    return name


def dft_spectrum(time_series, dt, freqs):
    """Computes the frequency spectrum associated to a time series directly
    using the discrete fourier transform, which may become slow for many
    frequencies and time points.

    Parameters
    ----------
    time_series: array_like
        1D array of time-dependent data.
    dt : float, optional
        Step in time over which the time series is recorded.
    freqs : array_like
        Array of frequencies to sample the spectrum at.

    Returns
    -------
    spectrum : array_like
        Array of same size as ``freqs`` giving the complex-valued spectrum.
    """

    frs = np.array(freqs)
    tdep = np.array(time_series)
    tmesh = np.arange(tdep.size) * dt
    spectrum = np.sum(
        tdep[:, np.newaxis]
        * np.exp(2j * np.pi * frs[np.newaxis, :] * tmesh[:, np.newaxis]),
        0,
    ).ravel()

    return dt / np.sqrt(2 * np.pi) * spectrum


def x_to_center(Ex):
    """Interpolate Ex positions to the center of a Yee lattice"""
    return (
        Ex
        + np.roll(Ex, -1, 1)
        + np.roll(Ex, -1, 2)
        + np.roll(np.roll(Ex, -1, 1), -1, 2)
    ) / 4


def y_to_center(Ey):
    """Interpolate Ey positions to the center of a Yee lattice"""
    return (
        Ey
        + np.roll(Ey, -1, 0)
        + np.roll(Ey, -1, 2)
        + np.roll(np.roll(Ey, -1, 0), -1, 2)
    ) / 4


def z_to_center(Ez):
    """Interpolate Ez positions to the center of a Yee lattice"""
    return (
        Ez
        + np.roll(Ez, -1, 0)
        + np.roll(Ez, -1, 1)
        + np.roll(np.roll(Ez, -1, 0), -1, 1)
    ) / 4


def E_to_center(Ex, Ey, Ez):
    return (x_to_center(Ex), y_to_center(Ey), z_to_center(Ez))


def eps_to_center(eps_xx, eps_yy, eps_zz):
    """Interpolate eps_r to the center of the Yee lattice. Used for plotting."""
    # # Simple averaging of one x, y, z values per cell.
    # return (eps_xx + eps_yy + eps_zz)/3

    # Average all 4 eps_xx, 4 eps_yy, and 4 eps_zz values around the
    # cell center, similar to the monitor field recording.
    return (
        x_to_center(eps_xx) + y_to_center(eps_yy) + z_to_center(eps_zz)
    ) / 3


def cs2span(center, size):
    """Get shape (3, 2) span from center and size, each (3, ) arrays."""
    return np.vstack(
        (
            np.array(center) - np.array(size) / 2,
            np.array(center) + np.array(size) / 2,
        )
    ).T


def span2cs(span):
    """Get center, size: each arrays of shape (3,), from a shape (3, 2) array
    span.
    """
    center = np.array([(span[d, 1] + span[d, 0]) / 2 for d in range(3)])
    size = np.array([(span[d, 1] - span[d, 0]) for d in range(3)])
    return center, size


def check_3D_lists(**kwargs):
    """ Verify that input arguments are lists with three elements """
    for key, val in kwargs.items():
        try:
            if not isinstance(val, list) and not isinstance(val, tuple):
                raise ValueError
            if len(val) != 3:
                raise ValueError
            for v in val:
                if type(v) in [list, tuple, np.ndarray]:
                    raise ValueError
        except:
            raise ValueError(
                "'%s' must be array-like with three elements." % key
            )


def axes_handed(axes):
    """Return +1 if the axes indexes in the list ``axes`` form a right-handed
    coordinate system, and -1 if they form a left-handed one.
    """

    if listify(axes) in [[0, 1, 2], [1, 2, 0], [2, 0, 1]]:
        return 1
    elif listify(axes) in [[0, 2, 1], [1, 0, 2], [2, 1, 0]]:
        return -1
    else:
        raise ValueError("Unrecognized list of axes indexes.")


def poynting_flux(E, H):
    """Compute the time-averaged Poynting vector that gives the energy
    flow per unit area per unit time at every point. ``E`` and ``H`` are
    assumed to be arrays of the same shape, as returned by frequency
    monitors. The first dimension is the field polarization (x, y, z), and
    must have size 3.
    """

    if E.shape != H.shape:
        raise ValueError("E and H must have the same dimension.")
    if E.shape[0] != 3:
        raise ValueError("First dimension of fields must have size 3.")

    return 1 / 2 * np.real(np.cross(E, np.conj(H), axis=0))



import sys
import logging as lg


class LessThanFilter(lg.Filter):
    def __init__(self, exclusive_maximum, name=""):
        super(LessThanFilter, self).__init__(name)
        self.max_level = exclusive_maximum

    def filter(self, record):
        # non-zero return means we log this message
        return 1 if record.levelno < self.max_level else 0


class MoreThanFilter(lg.Filter):
    def __init__(self, exclusive_minimum, name=""):
        super(MoreThanFilter, self).__init__(name)
        self.min_level = exclusive_minimum

    def filter(self, record):
        # non-zero return means we log this message
        return 1 if record.levelno > self.min_level else 0


def _level_str(level):
    """String to logging level.
    """
    if level.lower() == "debug":
        return lg.DEBUG
    elif level.lower() == "info":
        return lg.INFO
    elif level.lower() == "warning":
        return lg.WARNING
    elif level.lower() == "error":
        return lg.ERROR
    elif level.lower() == "critical":
        return lg.CRITICAL
    else:
        raise ValueError(f"Unrecognized level '{level}'.")

def log_and_raise(err_msg, exc_type):
    """ Log an erorr and raise an exception.
    """
    lg.error(err_msg)
    raise exc_type(err_msg)

def logging_default():
    """ Set the default logging configuration. File logging is reset. 
    INFO and DEBUG level messages are redirected to stdout. WARNING and above 
    go to stderr as usual. 
    """
    logger = lg.getLogger()
    while logger.hasHandlers():
        logger.removeHandler(logger.handlers[0])
    logger.setLevel(lg.NOTSET)
    formatter = lg.Formatter("%(levelname)s: %(message)s")

    logging_handler_err = lg.StreamHandler(sys.stderr)
    logging_handler_err.setLevel(lg.WARNING)
    logging_handler_err.addFilter(MoreThanFilter(lg.INFO))
    logging_handler_err.setFormatter(formatter)
    logger.addHandler(logging_handler_err)

    logging_handler_out = lg.StreamHandler(sys.stdout)
    logging_handler_out.setLevel(lg.INFO)
    logging_handler_out.addFilter(LessThanFilter(lg.WARNING))
    logger.addHandler(logging_handler_out)


def logging_level(level):
    """Set the lowest severity level to print to the console, as per the 
    python logger. This sets the level for both stdout and stderr output.

    Parameters
    ----------
    level : str
        One of ``['debug', 'info', 'warning', 'error', 'critical']``.
    """
    logger = lg.getLogger()
    lg_level = _level_str(level)
    for handler in logger.handlers:
        handler.setLevel(lg_level)



def logging_file(fname, filemode="w", level='debug'):
    """Set a file to write log to, independently from the stdout and stderr 
    output chosen using :meth:`logging_level`.

    Parameters
    ----------
    fname : str
        File name to direct the output to.
    filemode : str, optional
        'w' or 'a', defining if the file should be overwritten or appended.
    level : str
        One of 'debug', 'info', 'warning', 'error', 'critical'. This is 
        set for the file independently of the console output level set by 
        :meth:`logging_level`.
    """

    logger = lg.getLogger()
    file_handler = lg.FileHandler(fname, filemode)
    file_handler.setLevel(_level_str(level))
    formatter = lg.Formatter("%(levelname)s: %(message)s")
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)



from .structure import Structure, Box, Sphere, Cylinder, PolySlab, GdsSlab
from .source import (Source, VolumeSource, PointDipole, PlaneSource,
						PlaneWave, ModeSource, SourceTime, GaussianPulse)
from .monitor import Monitor, TimeMonitor, FreqMonitor, ModeMonitor
from .grid import Grid
from .material import Medium
from .dispersion import DispersionModel, Sellmeier
from .utils.em import dft_spectrum, poynting_flux
from .utils.log import logging_default, logging_level, logging_file

from . import material
PEC = material.PEC()
PMC = material.PMC()

from .simulation import Simulation
from .constants import C_0, ETA_0, EPSILON_0, MU_0, inf

logging_default()


import boto3
import getpass
import hashlib
import os
import functools
import requests
from .config import Config

boto3.setup_default_session(region_name='us-east-1')


def email2username(email):
    return email.replace('@', '-at-')


def authentication_api(email, password):
    url = f"{Config.WEB_API_ENDPONT}/auth"
    headers = {"Application":"TIDY3D"}
    resp = requests.get(url, headers=headers, auth=(email, password))
    access = resp.json()['data']
    auth = access['auth']
    keys = access['user']

    userItems = keys.items()
    newKeys={}
    for key, val in userItems:
        newKeys[''.join(key[:1].upper() + key[1:])] = val
        newKeys[key] = val
    Config.auth = auth
    Config.user = newKeys

def getEmailPasswdAuthKey():
    credentialPath = os.path.expanduser('~/.tidy3d')
    if os.path.exists('{0}/{1}'.format(credentialPath, 'email')) and \
            os.path.exists('{0}/{1}'.format(credentialPath, 'passwd')):
        with open(os.path.join(credentialPath, 'email'), 'r') as f:
            email = f.read()
        with open(os.path.join(credentialPath, 'passwd'), 'r') as f:
            password = f.read()
        try:
            authentication_api(email, password)
            return
        except:
            print('Error: Failed to log in with existing user:', email)
            print()
            pass
    while True:
        email = input('enter your email registered at tidy3d: ')
        password = getpass.getpass("enter your password: ")
        salt = '5ac0e45f46654d70bda109477f10c299'
        password = hashlib.sha512(password.encode('utf-8') + salt.encode('utf-8')).hexdigest()
        try:
            authentication_api(email, password)
            break
        except:
            print('Error: Failed to log in with new username and password.')
            print()
            pass
    while True:
        login = input('Do you want to keep logged in on this machine ([Y]es / [N]o) ')
        if login == 'Y' or login == 'y':
            os.makedirs(credentialPath, exist_ok=True)
            with open(os.path.join(credentialPath, 'passwd'), 'w') as f:
                f.write(password)
            with open(os.path.join(credentialPath, 'email'), 'w') as f:
                f.write(email)
            break
        elif login == 'N' or login == 'n':
            os.makedirs(credentialPath, exist_ok=True)
            break
        else:
            print('Unknown response: {0}\n'.format(login))


def refreshToken(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        resp = func(*args, **kwargs)
        return resp

    return wrapper


getEmailPasswdAuthKey()



import numpy as np
import logging

from ..utils import axes_handed, log_and_raise
from ..constants import int_, float_, complex_, fp_eps, ETA_0, C_0
from ..mode import Mode, ModePlane, dot_product
from .Source import VolumeSource, ModeSource, PlaneWave, PlaneSource

class SourceData(object):
    """Class used internally to store data related to a Source added in a 
    specific Simulation.
    """

    def __init__(self, source):

        # Reference to the source that the data corresponds to
        self.source = source

        # Mesh normalization, set by Simulation
        self.mesh_norm = 1

        # Mesh in time and time dependence, set by Simulation
        self.tmesh = None
        self.time_dep = np.zeros((0, 2), dtype=float_)

        # A complex-valued source can be used to launch complex-valued 
        # eigenmodes. It includes a second space and time dependence 
        # corresponding to the imaginary part of the source.
        self.complex = False

        # Phase offset of the magnetic current M and the electric current J.
        self.phaseJ = 0
        self.phaseM = 0

        # Extra variables specific to ModeSource
        if isinstance(source, ModeSource):
            # Variables to store the mode plane, mode, and mode index
            self.mode_plane = ModePlane(source.span, source.norm_ind)
            self.mode = None
            if (isinstance(source, PlaneWave) or 
                            isinstance(source, PlaneSource)):
                # No choice in modes
                self.mode_ind = 0
            else:
                # Must be set using Simulation.set_mode(source, mode_ind)
                self.mode_ind = None

            # Phases accountint for the temporal and spatial offset between J 
            # and M, set by Simulation.
            self.phaseMJ_s = 0
            self.phaseMJ_t = 0

    def _mesh_norm(self, mesh_step):
        """Normalize source by ``mesh_step`` in a given direction, if the span 
        of the source in that direction is zero. This is to avoid changes in 
        radiated power when the spatial discretization is changed.
        """

        self.mesh_norm = 1
        for dim in range(3):
            if np.abs(self.source.span[dim, 1] -
                        self.source.span[dim, 0]) < fp_eps:
                self.mesh_norm /= mesh_step[dim]

    def _set_tdep(self, tmesh, phaseM=0.):
        """Compute the time-dependence of the source currents.
        
        Parameters
        ----------
        tmesh : array_like
            Array of shape (Nt, ) defining the mesh in time, in seconds.
        phaseM : float, optional
            Phase offset of the magnetic current M from the electric current J.
        """

        self.phaseM = phaseM
        tdJ = self.source.source_time._get_time(tmesh, self.phaseJ)
        tdM = self.source.source_time._get_time(tmesh, self.phaseM)

        self.tmesh = tmesh
        self.time_dep = np.stack((tdJ, tdM), axis=1).astype(float_)

        # Time dependence offset by pi/2 for complex-valued sources
        tdJ = self.source.source_time._get_time(tmesh, np.pi/2 + self.phaseJ)
        tdM = self.source.source_time._get_time(tmesh, np.pi/2 + self.phaseM)
        self.time_dep_im = np.stack((tdJ, tdM), axis=1).astype(float_)

    def _get_spectrum(self, freqs):
        return self.source.source_time._get_spectrum(freqs, self.tmesh)

    def _currents(self, src_inds):
        """Source currents (Jx, Jy, Jz, Mx, My, Mz) at grid locations 
        specified by the global grid indexes ``src_inds``.
        """

        if isinstance(self.source, VolumeSource):
            src_amps = np.zeros((src_inds.shape[0], 6), dtype=float_)
            src_amps[:, self.source.components==1] = \
                self.source.amplitude * self.mesh_norm

        elif isinstance(self.source, ModeSource):
            """ Currents for the eigenmode source are defined through the 
            fields of the mode through the equivalence principle. Only 
            tangential currents are nonzero. ``src_inds`` are the indexes in 
            the global grid, so to get the indexing in the 2D plane, we use 
            ``mode_plane.span_inds``, which gives the beginning and end 
            indexes of the plane in the global grid.
            """

            src_amps = np.zeros((src_inds.shape[0], 6), dtype=complex_)
            mplane = self.mode_plane

            # Indexing in the 2D plane. 
            inds1 = src_inds[:, mplane.cross_inds[0]] - \
                        mplane.span_inds[mplane.cross_inds[0], 0]
            inds2 = src_inds[:, mplane.cross_inds[1]] - \
                        mplane.span_inds[mplane.cross_inds[1], 0]

            # Electric current sources from magnetic field of the eigenmode.
            src_amps[:, mplane.cross_inds[0]] = \
                                    self.mode.H[1, inds1, inds2].ravel()
            src_amps[:, mplane.cross_inds[1]] = \
                                    -self.mode.H[0, inds1, inds2].ravel()

            # Magnetic current sources from electric field of the eigenmode.
            src_amps[:, 3 + mplane.cross_inds[0]] = \
                                    -self.mode.E[1, inds1, inds2].ravel()
            src_amps[:, 3 + mplane.cross_inds[1]] = \
                                    self.mode.E[0, inds1, inds2].ravel()
            
            # +/-1 depending on handedness of mode_plane axes
            ax_fact = axes_handed([mplane.cross_inds[0], mplane.cross_inds[1],
                                    mplane.norm_ind])
            # dir_ind is +/-1 for positive/negative direction
            src_amps[:, 3:] *= ax_fact*self.source.dir_ind
            # Overall source amplitude
            src_amps *= self.source.amplitude * self.mesh_norm
            # Extra correction because J and M are not co-localized
            src_amps /= np.sqrt(np.cos(self.phaseMJ_s))

        return src_amps

    def _compute_modes_plane_wave(self, Nmodes):
        """ Compute modes without matrix diagonalization.
        """

        source = self.source
        mplane = self.mode_plane

        eps2D = mplane.eps
        if np.amax(eps2D) - np.amin(eps2D) > 1e-5:
            log_and_raise(
                "'PlaneWave' is intended to be placed "
                "entirealy in a homogeneous material.",
                RuntimeError
            )

        if Nmodes > 1:
            log_and_raise(
                "Only a single mode can be computed for a "
                "'PlaneWave' the plane wave with a specified polarization.",
                RuntimeError
            )

        frequency = source.source_time.frequency

        """ Mode fields are defined w.r.t. ModePlane axes, which are 
        [cross_ind1, cross_ind2, norm_ind]. """
        N1, N2 = eps2D.shape
        neff = np.sqrt(np.amax(eps2D))
        E = np.zeros((3, N1, N2))
        H = np.zeros((3, N1, N2))
        if source.epol_ind < source.hpol_ind:
            # E is polarized along first cross-index
            E[0, :, :] = 1
            H[1, :, :] = neff/ETA_0
        else:
            # E is polarized along second cross-index
            E[1, :, :] = 1
            H[0, :, :] = -neff/ETA_0

        mplane.freqs = [frequency]
        mode = Mode(E, H, neff, keff=0)
        flux = dot_product((E, H), (E, H), mplane.mesh_step)
        mode.E /= np.sqrt(flux)
        mode.H /= np.sqrt(flux)
        mode.kvector = np.array([0, 0, source.dir_ind * mode.neff])
        mode.kvector *= 2 * np.pi * source.source_time.frequency / C_0
        mplane.modes = [[mode]]

    def _compute_modes_plane_source(self, Nmodes):
        """ Compute (oblique) wave mode without matrix diagonalization.
        """

        source = self.source
        mplane = self.mode_plane

        eps2D = mplane.eps
        if np.amax(eps2D) - np.amin(eps2D) > 1e-5:
            log_and_raise(
                "'PlaneSource' is intended to be placed "
                "entirealy in a homogeneous material.",
                RuntimeError
            )

        if Nmodes > 1:
            log_and_raise(
                "Only a single mode can be computed for a "
                "'PlaneSource' the plane wave with a specified polarization.",
                RuntimeError
            )

        frequency = source.source_time.frequency

        """ Mode fields are defined w.r.t. ModePlane axes, which are 
        [cross_ind1, cross_ind2, norm_ind]. """
        N1, N2 = eps2D.shape
        neff = np.sqrt(np.amax(eps2D))
        E = np.zeros((3, N1, N2), dtype=complex_)
        H = np.zeros((3, N1, N2), dtype=complex_)

        # kvector in units of effective index, based on diffraction order
        d1 = N1 * mplane.mesh_step[0]
        d2 = N2 * mplane.mesh_step[1]
        k1 = source.diff_order[0] * 2*np.pi / d1 
        k2 = source.diff_order[1] * 2*np.pi / d2 
        k0 = 2 * np.pi * neff * frequency / C_0
        if k0 ** 2 <= k1**2 + k2**2:
            raise ValueError("Diffraction order [%d, %d] not available"%(
                            source.diff_order[0], source.diff_order[1]) +\
                            "for PlaneSource size and frequency.")
        kn = np.sqrt(k0**2 - k1**2 - k2**2)
        # Unit vector along the propagation axis
        dir_vec = np.array([k1, k2, kn])/k0
        # k-vector in 1/micron
        kvector = source.dir_ind * k0 * dir_vec

        # Unit vector along axis normal to injection plane
        norm_vec = np.array([0, 0, 1])

        if np.linalg.norm(source.polarization) < fp_eps:
            # Unit vector for S polarization
            # Add a tiny bit of offset to break S-P degeneracy at theta = 0
            dvec = np.copy(dir_vec)
            if np.linalg.norm(np.cross(norm_vec, dvec)) < fp_eps:
                dvec[0] += fp_eps
            S_vec = np.cross(norm_vec, dvec)
            S_vec /= np.sqrt(S_vec.dot(S_vec))

            # Unit vector for P polarization
            P_vec = np.cross(S_vec, dir_vec)
            P_vec /= np.sqrt(P_vec.dot(P_vec))

            # E-field polarization vector
            E_vec = np.cos(source.pol_angle)*P_vec +\
                        np.sin(source.pol_angle)*S_vec
            # H-field polarization vector
            H_vec = - np.sin(source.pol_angle)*P_vec +\
                        np.cos(source.pol_angle)*S_vec
        else:
            E_vec = source.polarization[[mplane.cross_inds[0],
                                         mplane.cross_inds[1],
                                         mplane.norm_ind]]
            E_vec -= np.dot(E_vec, dir_vec) * dir_vec
            E_vec /= np.sqrt(E_vec.dot(E_vec))
            H_vec = np.cross(dir_vec, E_vec)
            H_vec /= np.sqrt(H_vec.dot(H_vec))

        # Add impedence
        H_vec *= neff/ETA_0

        """ Finally need to account for the Yee grid locations of each component
        through mplane.mesh. E[0] and H[1] are co-locallized in the 
        cross-section plane, and same for E[1] and H[0]. The offset in the 
        normal direction is taken into account later. """
        xg, yg = np.meshgrid(mplane.mesh_e1[0],
                             mplane.mesh_e1[1])
        E[0, :, :] = E_vec[0]*np.exp(1j*(xg*kvector[0] + yg*kvector[1])).T
        H[1, :, :] = H_vec[1]*np.exp(1j*(xg*kvector[0] + yg*kvector[1])).T
        xg, yg = np.meshgrid(mplane.mesh_e2[0],
                             mplane.mesh_e2[1])
        E[1, :, :] = E_vec[1]*np.exp(1j*(xg*kvector[0] + yg*kvector[1])).T
        H[0, :, :] = H_vec[0]*np.exp(1j*(xg*kvector[0] + yg*kvector[1])).T

        if np.linalg.norm(np.cross(norm_vec, dir_vec)) > 2*fp_eps:
            self.complex = True

        mplane.freqs = [frequency]
        mode = Mode(E, H, neff, keff=0)
        flux = dot_product((E, H), (E, H), mplane.mesh_step)
        mode.E /= np.sqrt(flux)
        mode.H /= np.sqrt(flux)
        mode.kvector = kvector
        mplane.modes = [[mode]]

    def _compute_modes_mode_source(self, Nmodes):
        """ Compute general eigenmodes using matrix diagonalization.
        """
        frequency = self.source.source_time.frequency
        self.mode_plane.compute_modes(Nmodes, frequency)


import numpy as np

from . import log_and_raise

def poynting_flux(E, H):
    """Compute the time-averaged Poynting vector that gives the energy
    flow per unit area per unit time at every point. ``E`` and ``H`` are
    assumed to be arrays of the same shape, as returned by frequency
    monitors. The first dimension is the field polarization (x, y, z), and
    must have size 3.
    """

    if E.shape != H.shape:
        log_and_raise(
            "E and H must have the same dimension.",
            ValueError
        )
    if E.shape[0] != 3:
        log_and_raise(
            "First dimension of fields must have size 3.",
            ValueError
        )

    return 1 / 2 * np.real(np.cross(E, np.conj(H), axis=0))


def dft_spectrum(time_series, dt, freqs):
    """Computes the frequency spectrum associated to a time series directly
    using the discrete fourier transform.

    Note
    ----
    The DFT spectrum can be computed over an arbitrary list of frequencies, 
    but is much more inefficient than FFT. Use sparingly.

    Parameters
    ----------
    time_series: array_like
        1D array of time-dependent data.
    dt : float, optional
        Step in time over which the time series is recorded.
    freqs : array_like
        Array of frequencies to sample the spectrum at.

    Returns
    -------
    spectrum : array_like
        Array of same size as ``freqs`` giving the complex-valued spectrum.
    """

    frs = np.array(freqs)
    tdep = np.array(time_series)
    tmesh = np.arange(tdep.size) * dt
    spectrum = np.sum(
        tdep[:, np.newaxis]
        * np.exp(2j * np.pi * frs[np.newaxis, :] * tmesh[:, np.newaxis]),
        0,
    ).ravel()

    return dt / np.sqrt(2 * np.pi) * spectrum


def x_to_center(Ex):
    """Interpolate Ex positions to the center of a Yee lattice"""
    return (
        Ex
        + np.roll(Ex, -1, 1)
        + np.roll(Ex, -1, 2)
        + np.roll(np.roll(Ex, -1, 1), -1, 2)
    ) / 4


def y_to_center(Ey):
    """Interpolate Ey positions to the center of a Yee lattice"""
    return (
        Ey
        + np.roll(Ey, -1, 0)
        + np.roll(Ey, -1, 2)
        + np.roll(np.roll(Ey, -1, 0), -1, 2)
    ) / 4


def z_to_center(Ez):
    """Interpolate Ez positions to the center of a Yee lattice"""
    return (
        Ez
        + np.roll(Ez, -1, 0)
        + np.roll(Ez, -1, 1)
        + np.roll(np.roll(Ez, -1, 0), -1, 1)
    ) / 4


def E_to_center(Ex, Ey, Ez):
    return (x_to_center(Ex), y_to_center(Ey), z_to_center(Ez))


def eps_to_center(eps_xx, eps_yy, eps_zz):
    """Interpolate eps_r to the center of the Yee lattice.
    """
    
    # # Simple averaging of one x, y, z values per cell.
    # return (eps_xx + eps_yy + eps_zz)/3

    # Average all 4 eps_xx, 4 eps_yy, and 4 eps_zz values around the
    # cell center, similar to the monitor field recording.
    return (
        x_to_center(eps_xx) + y_to_center(eps_yy) + z_to_center(eps_zz)
    ) / 3


import requests

from .authentication import getEmailPasswdAuthKey
from .config import Config

class FileDoesNotExist(Exception):
    pass


def handle_response(func):
    def wrapper(*args, **kwargs):
        resp = func(*args, **kwargs)
        if resp.status_code == 401 and Config.auth_retry < 1:
            getEmailPasswdAuthKey()
            Config.auth_retry = Config.auth_retry + 1
            resp = func(*args, **kwargs)
        elif resp.status_code != 200:
            resp.raise_for_status()
        try:
            Config.auth_retry = 0
            jsn = resp.json()['data']
        except Exception as e:
            print('Could not json decode response : {0}!'.format(resp.text))
            raise
        return jsn

    return wrapper


@handle_response
def post2(method, data=None):
    queryUrl = f"{Config.WEB_API_ENDPONT}/{method}"
    headers = {'Authorization': f"Bearer {Config.auth['accessToken']}", 'FLOW360USER': Config.user['identityId'], "Application":"TIDY3D"}
    return requests.post(queryUrl, headers=headers, json=data)


@handle_response
def put2(method, data):
    queryUrl = f"{Config.WEB_API_ENDPONT}/{method}"
    headers = {'Authorization': f"Bearer {Config.auth['accessToken']}", 'FLOW360USER': Config.user['identityId'], "Application":"TIDY3D"}
    return requests.put(queryUrl, headers=headers, json=data)


@handle_response
def get2(method):
    queryUrl = f"{Config.WEB_API_ENDPONT}/{method}"
    headers = {'Authorization': f"Bearer {Config.auth['accessToken']}", 'FLOW360USER': Config.user['identityId'], "Application":"TIDY3D"}
    return requests.get(queryUrl, headers=headers)


@handle_response
def delete2(method):
    queryUrl = f"{Config.WEB_API_ENDPONT}/{method}"
    headers = {'Authorization': f"Bearer {Config.auth['accessToken']}", 'FLOW360USER': Config.user['identityId'], "Application":"TIDY3D"}
    return requests.delete(queryUrl, headers=headers)



import json
import time
import sys
import os
import requests
from datetime import datetime

from s3transfer import TransferConfig

from .config import Config
from .httputils import post2, get2, delete2, put2
from .s3utils import s3Client, keys, UploadProgress

from ..utils import log_and_raise


def new_project(
    sim_dict,
    task_name=None,
    folder_name="default",
    solver_version=Config.SOLVER_VERSION,
):
    """Create new project.

    Parameters
    ----------
    sim_dict : dict
        Dictionary containing all simulation parameters, which can be
        generated by :meth:`.Simulation.export`.
    task_name : str, optional
        Custom name for the task.
    folder_name: str, optional
        folder to hold the task.
    solver_version : str, optional
        Version of the Tidy3D solver.

    Returns
    -------
    dict
        Project dictionary with simulation and task data.
    """
    if task_name == None:
        task_name = f"Task_{int(datetime.now().timestamp())}"

    # Initialize the task
    try:
        task = post2(
            f"fdtd/model/{folder_name}/task",
            {
                "status": "draft",
                "solverVersion": solver_version,
                "taskName": task_name,
                "nodeSize": int(sim_dict["parameters"]["nodes"]),
                "timeSteps": int(sim_dict["parameters"]["time_steps"]),
                "computeWeight": float(
                    sim_dict["parameters"]["compute_weight"]
                ),
            },
        )
    except requests.exceptions.HTTPError as e:
        error_json = json.loads(e.response.text)
        log_and_raise(error_json["error"], RuntimeError)

    # upload the file to s3
    print("Uploading the json file...")
    resp = s3Client.put_object(
        Body=json.dumps(sim_dict, indent=4),
        Bucket=Config.STUDIO_BUCKET,
        Key="users/{0}/{1}/{2}".format(
            keys["UserId"], task["taskId"], "fdtd3d.json"
        ),
    )
    task["status"] = "queued"
    return put2(f'fdtd/model/{folder_name}/task/{task["taskId"]}', task)


def get_project(task_id):
    """Get all project details from a given task_id.

    Parameters
    ----------
    task_id : str
        Task identification string.

    Returns
    -------
    dict
        Project dictionary containing task data.
    """

    return get2(f"fdtd/task/{task_id}")


def get_projects():
    """Get a list with all details of all projects of the current user.

    Returns
    -------
    list
        A list of ``dict``, with one entry for every folder name. Each of
        those can contain multiple projects whose corresponding data is
        listed in the ``'children'`` entry.
    """

    return get2(f"fdtd/models")


def delete_project(task_id):
    """Delete a project from a given task_id.

    Parameters
    ----------
    task_id : str
        Task identification string.

    Returns
    -------
    dict
        Project dictionary of the deleted project.
    """

    return delete2(f"fdtd/task/{task_id}")


def delete_queued():
    """Delete all queued projects.

    Returns
    -------
    int
        Total number of projects deleted.
    """

    projects = get_projects()
    count = 0
    for pfolder in projects:
        for task in pfolder["children"]:
            if task["status"] == "queued":
                delete_project(task["taskId"])
                count += 1

    return count

def delete_old(days_old=100, folder=None):
    """Delete all projects older than a given amount of days.
    
    Parameters
    ----------
    days_old : int, optional
        Minimum number of days since the project creation.
    folder : str or None, optional
        If None, all folders are purged.
    
    Returns
    -------
    int
        Total number of projects deleted.
    """

    projects = get_projects()
    count = 0
    for pfolder in projects:
        if pfolder['name'] == folder or folder is None:
            for task in pfolder["children"]:
                stime_str = task["submitTime"]
                stime = datetime.strptime(stime_str, "%Y:%m:%d:%H:%M:%S")
                days_elapsed = (datetime.utcnow() - stime).days
                if days_elapsed >= days_old:
                    delete_project(task["taskId"])
                    count += 1

    return count


def monitor_project(task_id):
    """Monitor the status of a given project every second. Function exits when
    the exit status is either ``'success'`` or ``'error'``.

    Parameters
    ----------
    task_id : str
        Task identification string.
    """
    status = ""
    while status not in ["success", "error"]:
        project = get_project(task_id)
        status = project["status"]
        name = project["taskName"]
        sys.stdout.write('\rProject "' + name + '" status: %s       ' % status)
        sys.stdout.flush()
        time.sleep(0.5)
        sys.stdout.write('\rProject "' + name + '" status: %s.      ' % status)
        sys.stdout.flush()
        time.sleep(0.5)
        sys.stdout.write('\rProject "' + name + '" status: %s..     ' % status)
        sys.stdout.flush()
        time.sleep(0.5)
        sys.stdout.write('\rProject "' + name + '" status: %s...    ' % status)
        sys.stdout.flush()
        time.sleep(0.5)
    print("\n")


def get_last_projects(Nprojects=None):
    """Get a list with information about the last ``Nprojects`` submitted
    projects.

    Returns
    -------
    list
        A list of ``dict``, with one entry for every project.
    """

    projects = get_projects()
    store_dict = {
        "submit_time": [],
        "status": [],
        "task_name": [],
        "task_id": [],
    }
    for pfolder in projects:
        for task in pfolder["children"]:
            try:
                store_dict["submit_time"].append(task["submitTime"])
                store_dict["status"].append(task["status"])
                store_dict["task_name"].append(task["taskName"])
                store_dict["task_id"].append(task["taskId"])
            except:
                pass

    sort_inds = sorted(
        range(len(store_dict["submit_time"])),
        key=store_dict["submit_time"].__getitem__,
        reverse=True,
    )

    if Nprojects is None or Nprojects > len(sort_inds):
        Nprojects = len(sort_inds)

    out_dict = []
    for ipr in range(Nprojects):
        out_dict.append(
            {key: item[sort_inds[ipr]] for (key, item) in store_dict.items()}
        )

    return out_dict


def list_projects(Nprojects=None):
    """Print a summary of all projects of the current user, in chronological
    order of submission.

    Parameters
    ----------
    Nprojects : None, optional
        If supplied, only the last ``Nprojects`` will be listed.
    """

    projects_info = get_last_projects(Nprojects)

    for pr_info in projects_info:
        print("Project name: %s" % pr_info["task_name"])
        print("Task ID     : %s" % pr_info["task_id"])
        print("Submit time : %s" % pr_info["submit_time"])
        print("Status      : %s" % pr_info["status"])
        print("-" * 50)


def download_results_file(task_id, src, target=None):
    """Download all the results recorded by simulation monitors.

    Parameters
    ----------
    task_id : str
        Task ID of the project (after a successful run).
    src : str
        File to download. Valid file names are ``'monitor_data.hdf5'`` and
        ``'tidy3d.log'``.
    target : None, optional
        Filename to store locally. If ``None``, the same name as ``src``
        will be used.
    """
    if target is None:
        target = src
    if src is None:
        log_and_raise("'src' cannot be None.", ValueError)

    try:
        project = get_project(task_id)
    except:
        log_and_raise("Unable to get project " + task_id, RuntimeError)

    try:
        s3Client.download_file(
            Bucket=Config.STUDIO_BUCKET,
            Filename=target,
            Key="users/{0}/{1}/output/{2}".format(
                keys["UserId"], task_id, src
            ),
        )
    except:
        log_and_raise(
            "Cannot retrieve requested file, check the file name and "
            "make sure the project has run correctly. Current "
            f"project status is '{project['status']}'.",
            RuntimeError
        )
        
def download_results(task_id, target_folder=""):
    """Download the results of a solver run, including the json file defining
    the simulation, a single file containing all the monitor data, and a log
    file.

    Parameters
    ----------
    task_id : str
        Task ID of the project (after a successful run).
    target_folder : str, optional
        Folder in which to download the files. Default is working folder.
    """

    if target_folder != "" and not os.path.exists(target_folder):
        os.makedirs(target_folder)

    for fname in ["monitor_data.hdf5", "tidy3d.log"]:
        download_results_file(
            task_id, fname, target=os.path.join(target_folder, fname)
        )

    download_json(task_id, target_folder=target_folder)


def download_json(task_id, target_folder=""):
    """Download the project json file

    Parameters
    ----------
    task_id : str
        Task ID of the project (after a successful run).
    target_folder : str, optional
        Folder in which to download the files. Default is working folder.
    """

    if target_folder != "" and not os.path.exists(target_folder):
        os.makedirs(target_folder)

    project = get_project(task_id)
    sim_dict = project["taskParam"]
    if not sim_dict:
        s3Client.download_file(
            Bucket=Config.STUDIO_BUCKET,
            Filename=os.path.join(target_folder, "simulation.json"),
            Key="users/{0}/{1}/{2}".format(
                keys["UserId"], task_id, "fdtd3d.json"
            ),
        )
    else:
        with open(
            os.path.join(target_folder, "simulation.json"), "w"
        ) as sim_file:
            json.dump(json.loads(), sim_file, indent=4)



import numpy as np
import logging

from ..constants import int_, float_, complex_, fp_eps, ETA_0, C_0
from ..mode import dot_product, Mode
from ..utils import log_and_raise

from .Source import VolumeSource, ModeSource, PlaneWave, PlaneSource

def _compute_modes_source(self, source, Nmodes):
    src_data = self._source_ids[id(source)]
    if isinstance(source, PlaneSource):
        src_data._compute_modes_plane_source(Nmodes)
    elif isinstance(source, PlaneWave):
        src_data._compute_modes_plane_wave(Nmodes)
    elif isinstance(source, ModeSource):
        src_data._compute_modes_mode_source(Nmodes)

def spectrum(self, source, freqs):
    """Returns the spectrum of a :class:`.Source`.
    
    Parameters
    ----------
    source : Source
        Source object.
    freqs : array_like
        (Hz) Array of frequencies to evaluate the spectrum over.
    """
    src_data = self._source_ids[id(source)]
    return src_data._get_spectrum(freqs)

def set_mode(self, source, mode_ind):
    """Set the mode to be used as a source from the list of modes pre-computed 
    with :meth:`.compute_modes`.
    
    Parameters
    ----------
    mode_ind : int
        Index of the mode to use. Use :meth:`.viz_modes` to visualize the 
        modes and select the index to set.
    """

    src_data = self._source_ids[id(source)]
    mplane = src_data.mode_plane

    if isinstance(source, PlaneWave) or isinstance(source, PlaneSource):
        if mode_ind > 0:
            log_and_raise(
                "Plane wave mode index can only be set to 0, "
                "denoting the plane wave with a specified polarization.",
                RuntimeError
            )

        if src_data.mode is None:
            self.compute_modes(source, 1)
        src_data.mode = mplane.modes[0][mode_ind]
        src_data.mode_ind = 0

    elif isinstance(source, ModeSource):
        
        # Compute modes if selected mode index not currently stored
        if mode_ind >= len(mplane.modes[0]):
            logging.info(
                "Selected mode index larger than number of "
                "stored modes; recomputing modes up to selected index."
                )
            self.compute_modes(source, mode_ind + 1)

        # Copy selected mode dictionary
        mode = mplane.modes[0][mode_ind].to_dict()

        # Check that mode is localized
        Eedge = np.sum(np.abs(mode['E'][:, 0, :])**2 +
                        np.abs(mode['E'][:, -1, :])**2) + \
                np.sum(np.abs(mode['E'][:, :, 0])**2 +
                        np.abs(mode['E'][:, :, -1])**2)
        Enorm = np.sum(np.abs(mode['E'])**2)

        if Eedge/Enorm > 1e-3:
            logging.warning(
                "Mode field does not decay at boundaries, "
                "which may produce unexpected results."
                )

        for key in ['E', 'H']:
            max_im = np.amax(np.abs(np.imag(mode[key])))
            if max_im > 1e-5:
                logging.warning(
                    f"Ignoring imaginary part of field {key} "
                    f"with maximum magnitude {max_im:.4f}"
                    )
            mode[key] = np.real(mode[key])

        src_data.mode = Mode(mode['E'], mode['H'], mode['neff'], mode['keff'])
        neff_z = source.dir_ind*src_data.mode.neff
        src_data.mode.kvector = np.array([0, 0, neff_z])
        src_data.mode.kvector *= 2*np.pi * source.source_time.frequency / C_0
        src_data.mode_ind = mode_ind

    # Account for phase offset between M and J sources because of 
    # spatial and temporal mismatch.
    # Temporal offset phase
    src_data.phaseMJ_t = np.pi * source.source_time.frequency * \
                        src_data.mode_plane.time_step
    # Spatial offset phase
    src_data.phaseMJ_s = - src_data.mode.kvector[2]/2 * \
                        src_data.mode_plane.mesh_step[2]
    phaseM = src_data.phaseMJ_t + src_data.phaseMJ_s
    src_data._set_tdep(self.grid.tmesh, phaseM)


import numpy as np
import logging

from ..constants import int_, float_, complex_, fp_eps, xyz_dict, xyz_list
from ..utils.em import poynting_flux
from ..utils import log_and_raise
from ..mode import dot_product

from .Monitor import ModeMonitor

def _compute_modes_monitor(self, mode_object, Nmodes):
    """Compute the eigenmodes of the 2D cross-section of a ModeMonitor object.
    """   

    mnt_data = self._monitor_ids[id(mode_object)]
    mnt_data.mode_plane.compute_modes(Nmodes, mnt_data.freqs)

def data(self, monitor):
    """Return a dictionary with all the stored data in a :class:`.Monitor`.
    
    Parameters
    ----------
    monitor : Monitor
        The queried monitor.
    
    Returns
    -------
    dict
        Dictonary with all the set data fields in the monitor. For example, in 
        a frequency monitor, after a simulation run, ``dict['E']`` and 
        ``dict['H']`` are 5D arrays of shape ``(3, Nx, Ny, Nz, Nf)`` The first 
        index is the vector component of the field, while the next four 
        dimensions index the x, y, z, position index and the frequency index 
        at which the data was stored.
    """
    mnt_data = self._monitor_ids[id(monitor)]

    data_dict = {
                'E': mnt_data.E,
                'H': mnt_data.H,
                'S': mnt_data.S,
                'xmesh': mnt_data.xmesh,
                'ymesh': mnt_data.ymesh,
                'zmesh': mnt_data.zmesh,
                'tmesh': mnt_data.tmesh,
                'freqs': mnt_data.freqs,
                }

    return_dict = {}

    for key, val in data_dict.items():
        if val.size > 0:
            return_dict[key] = val

    return return_dict

def poynting(self, monitor):
    """Compute the Poynting vector at every point recorded by a 
    :class:`.Monitor`. Returns the instantaneous power flux per unit area at 
    every time for a :class:`.TimeMonitor`, and the time-averaged power flux 
    per unit area at every frequency for a :class:`.FreqMonitor`.
     
    Returns
    -------
    np.ndarray
        (Watt/micron\ :sup:`2`) The Poynting vector, i.e. the directed power 
        flux per unit area, at every sampling point of the monitor. Same shape 
        as the ``E`` and ``H`` fields stored in the monitor.
    """

    mnt_data = self._monitor_ids[id(monitor)]

    if mnt_data.E.size==0:
        log_and_raise(
            "No electric field stored in the monitor.",
            ValueError
        )
    if mnt_data.H.size==0:
        log_and_raise(
            "No magnetic field stored in the monitor.",
            ValueError
        )

    mnt_data.S = poynting_flux(mnt_data.E, mnt_data.H)
    return mnt_data.S

def flux(self, monitor, normal=None):
    """Compute the area-integrated Poynting flux in a given direction. 
    This is the total power flowing through a plane orthogonal to the 
    ``normal`` direction. If the monitor is larger than one in that 
    direction, the flux at every pixel is returned. Returns the instantaneous 
    power flux at every time for a :class:`.TimeMonitor`, and the 
    time-averaged power flux at every frequency for a :class:`.FreqMonitor`.
    
    Parameters
    ----------
    normal : {'x', 'y', 'z'}, or None
        If ``None``, normal is set to the first dimension along which the 
        Monitor spans a single pixel.
    
    Returns
    -------
    np.ndarray
        (Watt) The Poynting flux, an array of shape ``(Nsample, Np)``, where 
        ``Np`` are the number of points in the monitor volume along the 
        normal direction, while ``Nsample`` is the number of time steps 
        or frequencies in the monitor.
    """

    mnt_data = self._monitor_ids[id(monitor)]

    if normal is None:
        dmin = np.argmin(mnt_data.inds_end - mnt_data.inds_beg)
        normal = xyz_list[dmin]
    try:
        norm_ind = xyz_dict[normal]
    except:
        log_and_raise("'normal' must be one of 'x', 'y', or 'z'.", ValueError)

    cross_inds = [0, 1, 2]
    cross_inds.pop(norm_ind)

    if mnt_data.S.size==0:
        self.poynting(monitor)

    # Compute numerical integral.
    fl = np.prod(mnt_data.mesh_step[cross_inds]) * \
                np.sum(mnt_data.S[norm_ind, :, :, :, :],
                    axis=(cross_inds[0], cross_inds[1]))

    # Transpose so that time/frequency index comes first
    return fl.T

def decompose(self, mode_monitor, Nmodes=None):
    """Compute the decomposition of the fields recorded in a 
    :class:`.ModeMonitor` into the eigenmodes in the monitor plane.  
    
    Parameters
    ----------
    Nmodes : int or None, optional
        If None, the modes from the last call of :meth:`.compute_modes` will 
        be used.
    
    Returns
    -------
    np.ndarray
        A tuple of two arrays giving the overlap coefficients of the mode 
        fields with the forward- and backward-propagating eigenmodes, 
        respectively. Each array has shape ``(Nfreqs, Nmodes)``, where 
        ``Nfreqs`` is the number of frequencies in the monitor.
    """

    if not isinstance(mode_monitor, ModeMonitor):
        log_and_raise("'ModeMonitor' instance expected.", TypeError)

    mnt_data = self._monitor_ids[id(mode_monitor)]

    if Nmodes is not None:
        self.compute_modes(mode_monitor, Nmodes)

    if len(mnt_data.mode_plane.modes[0])==0:
        log_and_raise(
            "No modes found in monitor, use 'Simulation.compute_modes()' "
            "first.",
            RuntimeError
        )

    if mnt_data.data is False:
        log_and_raise(
            "Solver results not loaded into monitor yet.",
            RuntimeError
        )

    Nfreqs = len(mnt_data.freqs)
    Nmodes = len(mnt_data.mode_plane.modes[0])
    positive_coeff = np.zeros((Nfreqs, Nmodes), dtype=complex_)
    negative_coeff = np.zeros((Nfreqs, Nmodes), dtype=complex_)

    for ifreq in range(Nfreqs):
        # Need to get the monitor field of shape (3, Ncross1, Ncross2)
        E = mnt_data.E[mnt_data.cross_inds, :, :, :, ifreq]
        H = mnt_data.H[mnt_data.cross_inds, :, :, :, ifreq]
        fields_monitor = (np.squeeze(E, axis=1 + mnt_data.norm_ind),
                        np.squeeze(H, axis=1 + mnt_data.norm_ind))

        for imode, mode in enumerate(mnt_data.mode_plane.modes[ifreq]):
            # Fields of the mode snapped to Yee grid centers
            (Em, Hm) = mode.fields_to_center()
            # Overlap with positive-direction mode
            positive_coeff[ifreq, imode] = dot_product(
                                    (Em, Hm), fields_monitor,
                                    mnt_data.mode_plane.mesh_step)
            # Overlap with negative-direction mode.
            # Note: the sign of the fields is correct only for the tangential 
            # components, but those are the only ones entering the dot product.
            negative_coeff[ifreq, imode] = dot_product(
                                    (np.conj(Em), -np.conj(Hm)), 
                                    fields_monitor,
                                    mnt_data.mode_plane.mesh_step)

    return positive_coeff, negative_coeff


import numpy as np
import json
import h5py
import logging

from .utils import listify, list2str, object_name, cs2span, log_and_raise
from .utils.check import check_3D_lists
from .constants import int_, float_, fp_eps, C_0

from .grid import Grid
from .structure import Structure, Box
from .material import Medium
from . import PEC, PMC

from .source import Source, ModeSource, SourceData
from .monitor import Monitor, TimeMonitor, FreqMonitor, ModeMonitor
from .monitor import MonitorData

from .json_ops import write_parameters, write_structures, write_sources
from .json_ops import write_monitors

class Simulation(object):
    """
    Main class for building a simulation model.
    """
    from .utils.check import _check_size, _check_monitor_size
    from .source._simulation import _compute_modes_source, set_mode, spectrum
    from .monitor._simulation import _compute_modes_monitor, data, poynting
    from .monitor._simulation import flux, decompose
    from .json_ops import _read_simulation
    from .viz import _fmonitors_png, _structure_png
    from .viz import viz_eps_2D, viz_field_2D, viz_modes
    from .viz import viz_source, viz_source_spectrum, viz_source_time

    def __init__(self,
                size,
                center=[0., 0., 0.],
                resolution=None,
                mesh_step=None,
                structures=None,
                sources=None,
                monitors=None,
                symmetries=[0, 0, 0],
                pml_layers=[0, 0, 0],
                run_time=0.,
                courant=0.9
                ):
        """Construct.

        Parameters
        ----------
        center : array_like, optional
            (micron) 3D vector defining the center of the simulation domain.
        size : array_like, optional
            (micron) 3D vector defining the size of the simulation domain.
        resolution : float or array_like, optional
            (1/micron) Number of pixels per micron, or a 3D vector defining 
            the number of pixels per mircon in x, y, and z seprately.
        mesh_step : float or array_like, optional
            (micron) Step size in all directions, or a 3D vector defining the 
            step size in x, y, and z seprately. If provided, ``mesh_step`` 
            overrides the ``resolution`` parameter, otherwise 
            ``mesh_step = 1/resolution``.
        structures : Structure or List[Structure], optional
            Empty list (default) means vacuum. 
        sources : Source or List[Source], optional
            Source(s) to be added to the simulation.
        monitors : Monitor or List[Monitor], optional
            Monitor(s) to be added to the simulation.
        symmetries : array_like, optional
            Array of three integers defining reflection symmetry across a 
            plane bisecting the simulation domain normal to the x-, y-, and 
            z-axis, respectively. Each element can be ``0`` (no symmetry), 
            ``1`` (even, i.e. 'PMC' symmetry) or ``-1`` (odd, i.e. 'PEC' 
            symmetry). Note that the vectorial nature of the fields must be 
            taken into account to correctly determine the symmetry value.
        pml_layers : array_like, optional
            Array of three integers defining the number of PML layers on both 
            sides of the simulation domain along x, y, and z. When set to 
            ``0`` (default), periodic boundary conditions are applied.
        run_time : float, optional
            (second) Total electromagnetic evolution time.
        courant : float, optional
            Courant stability factor, must be smaller than 1, or more 
            generally smaller than the smallest refractive index in the 
            simulation.
        """

        check_3D_lists(center=listify(center), size=listify(size),
                            symmetries=listify(symmetries),
                            pml_layers=listify(pml_layers))

        logging.info("Initializing simulation...")

        # Set PML size. Npml defines the number of layers on all 6 sides.
        self.pml_layers = listify(pml_layers)
        self.Npml = np.vstack((pml_layers, pml_layers)).astype(int_).T

        # Set spatial mesh step
        if mesh_step is None:
            if resolution is None:
                log_and_raise(
                    "Either 'mesh_step' or 'resolution' must be set.",
                    ValueError
                )
            mesh_step = 1/np.array(resolution)
        else:
            if resolution is not None:
                logging.info(
                    "Note: parameter 'mesh_step' overrides 'resolution'."
                    )

        # Set simulation size inside the PML and including the PML
        self.center = np.array(center, dtype=float_)
        self.size_in = np.array(size, dtype=float_)
        self.size = self.size_in + 2*np.array(pml_layers)*mesh_step
        self.span = cs2span(self.center, self.size)

        # Initialize grid
        self.grid = Grid(self.span, mesh_step, symmetries, courant)
        logging.info(
            "Simulation domain in number of pixels: "
            f"{list2str(self.grid.Nxyz, '%d')}."
            )

        # Computational domain including symmetries, if any
        self.span_sym = np.copy(self.span)
        self.Nxyz_sym = np.copy(self.grid.Nxyz)
        for d, sym in enumerate(symmetries):
            if sym==-1:
                self.span_sym[d, 0] += self.size[d]/2
                self.Nxyz_sym[d] = self.Nxyz_sym[d]//2
            elif sym==1:
                self.span_sym[d, 0] += self.size[d]/2 - self.grid.mesh_step[d]
                self.span_sym[d, 1] += self.grid.mesh_step[d]
                self.Nxyz_sym[d] = self.Nxyz_sym[d]//2 + 2
        # Print new size, if there are any symmetries
        if np.any(np.array(symmetries)!=0):
            logging.info(
                "Computation domain (after symmetries): "
                f"{list2str(self.Nxyz_sym, '%d')}."
                )
        # Total number of points in computational domain (after symmetries)
        self.Np = np.prod(self.Nxyz_sym)
        # Print mesh step
        logging.info(
            f"Mesh step (micron): {list2str(self.grid.mesh_step, '%1.2e')}."
            )

        # Set up run time
        self.run_time = run_time
        self.courant = courant
        self.grid.set_tmesh(self.run_time)
        self.Nt = np.array(self.grid.tmesh.size)
        logging.info(f"Total number of time steps: {self.Nt}.")
        if self.Nt <= 0:
            logging.warning(
                f"run_time = {self.run_time:.2e} smaller than a single "
                f"simulation time step dt = {self.grid.dt:.2e}.",
            )

        # Check the simulation size
        self._check_size()

        # Materials and indexing populated when adding ``Structure`` objects.
        self._mat_inds = [] # material index of each structure
        self._materials = [] # list of materials included in the simulation
        self._structures = []

        # List containing SourceData for all sources, and a dictionary 
        # used to get SourceData from id(source), e.g. src_data = 
        # self._source_ids[id(source)]
        self._source_data = []
        self._source_ids = {}

        # List containing MonitorData for all monitors, and a dictionary 
        # used to get MonitorData from id(monitor)
        self._monitor_data = []
        self._monitor_ids = {}

        # Structures and material indexing for symmetry boxes
        self._structures_sym = [] # PEC/PMC boxes added for symmetry
        self._mat_inds_sym = []

        # Add structures, sources, monitors, symmetries
        if structures:
            self.add(structures)
        if sources:
            self.add(sources)
        if monitors:
            self.add(monitors)
        self._add_symmetries(symmetries)

        # JSON file from which the simulation is loaded
        self.fjson = None

    def __repr__(self):
        rep = "Tidy3D Simulation:\n"
        rep += "center      = %s\n" % list2str(self.center, "%1.4f")
        rep += "size        = %s\n" % list2str(self.size_in, "%1.4f")
        rep += "size w. pml = %s\n" % list2str(self.size, "%1.4f")
        rep += "mesh_step   = %s\n" % list2str(self.grid.mesh_step, "%1.4f")
        rep += "run_time    = %1.2e\n"%self.run_time
        rep += "symmetries  = %s\n" % list2str(self.symmetries, "%d")
        rep += "pml_layers  = %s\n\n" % list2str(self.pml_layers, "%d")

        rep += "Number of pixels in x, y, z: %s\n" % list2str(
                    self.grid.Nxyz, "%d")
        rep += "Number of pixels including \n" +\
               "symmetries                 : %s\n"%list2str(
                    self.Nxyz_sym, "%d")
        rep += "Number of nodes            : %d\n" % self.Np
        rep += "Number of time steps       : %d\n" % self.Nt
        rep += "Number of strictures       : %d\n"%len(self._structures)
        rep += "Number of sources          : %d\n"%len(self.sources)
        rep += "Number of monitors         : %d\n"%len(self.monitors)

        return rep

    @property
    def materials(self):
        """ List conaining all materials included in the simulation."""
        return self._materials

    @property
    def mat_inds(self):
        """ List conaining the material index in :attr:`.materials` of every 
        structure in :attr:`.structures`. """
        return self._mat_inds + self._mat_inds_sym

    @property
    def structures(self):
        """ List conaining all :class:`Structure` objects. """
        return self._structures + self._structures_sym

    @structures.setter
    def structures(self, new_struct):
        raise RuntimeError("Structures can be added upon Simulation init, "
                            "or using 'Simulation.add()'")

    @property
    def sources(self):
        """ List conaining all :class:`Source` objects. """
        return [src_data.source for src_data in self._source_data]

    @sources.setter
    def sources(self, new_sources):
        raise RuntimeError("Sources can be added upon Simulation init, "
                            "or using 'Simulation.add()'")

    @property
    def monitors(self):
        """ List conaining all :class:`.Monitor` objects. """
        return [mnt_data.monitor for mnt_data in self._monitor_data]

    @monitors.setter
    def monitors(self, new_monitors):
        raise RuntimeError("Monitors can be added upon Simulation init, "
                            "or using 'Simulation.add()'")

    def _add_structure(self, structure):
        """ Adds a Structure object to the list of structures and to the 
        permittivity array. """
        self._structures.append(structure)

        try:
            mind = self.materials.index(structure.material)
            self._mat_inds.append(mind)
        except ValueError:
            if len(self.materials) < 200:
                self._materials.append(structure.material)
                self._mat_inds.append(len(self.materials)-1)
            else:
                log_and_raise(
                    "Maximum 200 distinct materials allowed.",
                    RuntimeError
                )

    def _add_source(self, source):
        """ Adds a Source object to the list of sources.
        """

        if id(source) in self._monitor_ids.keys():
            logging.warning("Source already in Simulation, skipping.")
            return

        src_data = SourceData(source)
        src_data.name = object_name(self._source_data, source, 'source')
        src_data._mesh_norm(self.grid.mesh_step)
        src_data._set_tdep(self.grid.tmesh)
        self._source_data.append(src_data)
        self._source_ids[id(source)] = src_data

        if isinstance(source, ModeSource):
            src_data.mode_plane._set_eps(self)

    def _add_monitor(self, monitor):
        """ Adds a time or frequency domain Monitor object to the 
        corresponding list of monitors.
        """

        if id(monitor) in self._monitor_ids.keys():
            logging.warning("Monitor already in Simulation, skipping.")
            return

        mnt_data = MonitorData(monitor)
        mnt_data.name = object_name(self._monitor_data, monitor, 'monitor')
        self._monitor_data.append(mnt_data)
        self._monitor_ids[id(monitor)] = mnt_data

        memGB = self._check_monitor_size(monitor)
        logging.info(
            f"Estimated data size of monitor {mnt_data.name}: {memGB:.4f}."
            )

        # Initialize the ModePlane of a ModeMonitor
        if isinstance(monitor, ModeMonitor):
            mnt_data._set_mode_plane()
            mnt_data.mode_plane._set_eps(self)

    def _add_symmetries(self, symmetries):
        """ Add all symmetries as PEC or PMC boxes.
        """
        self.symmetries = listify(symmetries)
        for dim, sym in enumerate(symmetries):
            if sym not in [0, -1, 1]:
                log_and_raise(
                    "Reflection symmetry values can be 0 (no symmetry), "
                    "1, or -1.",
                    ValueError
                )
            elif sym==1 or sym==-1:
                sym_cent = np.copy(self.center)
                sym_size = np.copy(self.size)
                sym_cent[dim] -= self.size[dim]/2
                sym_size[dim] = sym_size[dim] + fp_eps
                sym_mat = PEC if sym==-1 else PMC
                sym_pre = 'pec' if sym==-1 else 'pmc'
                self._structures_sym.append(Box(center=sym_cent,
                                                size=sym_size,
                                                material=sym_mat,
                                                name=sym_pre + '_sym%d'%dim))
                try:
                    mind = self.materials.index(sym_mat)
                    self._mat_inds_sym.append(mind)
                except ValueError:
                    self._materials.append(sym_mat)
                    self._mat_inds_sym.append(len(self.materials)-1)

    def _pml_config(self):
        """Set the CPML parameters. Default configuration is hard-coded. This 
        could eventually be exposed to the user, or, better, named PML 
        profiles can be created.
        """
        cfs_config = {'sorder': 3, 'smin': 0., 'smax': None, 
                    'korder': 3, 'kmin': 1., 'kmax': 3., 
                    'aorder': 1, 'amin': 0., 'amax': 0}
        return cfs_config

    def _get_eps(self, mesh, edges='in', pec_val=-1e6, pmc_val=-3e6):
        """ Compute the permittivity over a given mesh. For large simulations, 
        this could be computationally heavy, so preferably use only over small 
        meshes (e.g. 2D cuts). 
        
        Parameters
        ----------
        mesh : tuple
            Three 1D arrays defining the mesh in x, y, z.
        edges : {'in', 'out', 'average'}
            When an edge of a structure sits exactly on a mesh point, it is 
            counted as in, out, or an average value of in and out is taken.
        pec_val : float
            Value to use for PEC material.
        pmc_val : float
            Value to use for PMC material.
        
        Returns
        -------
        eps : np.ndarray
            Array of size (mesh[0].size, mesh[1].size, mesh[2].size) defining 
            the relative permittivity at each point.
        """

        Nx, Ny, Nz = [mesh[i].size for i in range(3)]

        eps = np.ones((Nx, Ny, Nz), dtype=float_)

        # Apply all structures
        for struct in self.structures:

            eps_val = struct._get_eps_val(pec_val, pmc_val)
            struct._set_val(mesh, eps, eps_val, edges=edges)

        # return eps array after filling in all structures
        return eps

    def add(self, objects):
        """Add a list of objects, which can contain structures, sources, and 
        monitors.
        """

        for obj in listify(objects):
            if isinstance(obj, Structure):
                self._add_structure(obj)
            elif isinstance(obj, Source):
                self._add_source(obj)
            elif isinstance(obj, Monitor):
                self._add_monitor(obj)

    def load_results(self, dfile):
        """Load all monitor data recorded from a Tidy3D run.
        The data from each monitor can then be queried using 
        :meth:`.data`.
        
        Parameters
        ----------
        dfile : str
            Path to the file containing the simulation results.
        """

        mfile = h5py.File(dfile, "r")
        for (im, mnt_data) in enumerate(self._monitor_data):
            mname = mnt_data.name
            mnt_data._load_fields(mfile[mname]["indspan"][0, :],
                            mfile[mname]["indspan"][1, :],
                            np.array(mfile[mname]["E"]),
                            np.array(mfile[mname]["H"]), 
                            self.symmetries, self.grid.Nxyz)
            mnt_data.xmesh = np.array(mfile[mname]["xmesh"])
            mnt_data.ymesh = np.array(mfile[mname]["ymesh"])
            mnt_data.zmesh = np.array(mfile[mname]["zmesh"])
            mnt_data.mesh_step = self.grid.mesh_step

        mfile.close()
        fmonitors = [mnt for mnt in self.monitors
                            if isinstance(mnt, FreqMonitor)]
        if len(fmonitors) > 0:
            if len(self.sources) > 0: 
                self.source_norm(self.sources[0])
                logging.info(
                    "Applying source normalization to all frequency "
                    "monitors using source index 0.\nTo revert, "
                    "use Simulation.source_norm(None)."
                    )
                if len(self.sources) > 1:
                    logging.info(
                        "To select a different source for the normalization: "
                        "Simulation.source_norm(source)."
                        )
                        

    def source_norm(self, source):
        """Normalize all frequency monitors by the spectrum of a 
        :class:`.Source` object.
        
        Parameters
        ----------
        source : Source or None
            If ``None``, the normalization is reset to the raw field output.
        """

        if source is None:
            for mnt_data in self._monitor_data:
                mnt_data.set_source_norm(None)
            return

        src_data = self._source_ids[id(source)]

        for mnt_data in self._monitor_data:
            mnt_data.set_source_norm(src_data)

    def compute_modes(self, mode_object, Nmodes):
        """Compute the eigenmodes of the 2D cross-section of a 
        :class:`.ModeSource` or :class:`.ModeMonitor` object, assuming 
        translational invariance in the third dimension. The eigenmodes are 
        computed in decreasing order of propagation constant, at the central 
        frequency of the :class:`.ModeSource` or for every frequency in the 
        list of frequencies of the :class:`.ModeMonitor`. In-plane, periodic 
        boundary conditions are assumed, such that the mode shold decay at the 
        boundaries, or be matched with periodic boundary conditions in the 
        simulation. Use :meth:`.viz_modes` to visuzlize the computed 
        eigenmodes.
        
        Parameters
        ----------
        mode_object : ModeSource or ModeMonitor
            The object defining the 2D plane in which to compute the modes.
        Nmodes : int
            Number of eigenmodes to compute.
        """

        if isinstance(mode_object, Monitor):
            self._compute_modes_monitor(mode_object, Nmodes)
        elif isinstance(mode_object, Source):
            self._compute_modes_source(mode_object, Nmodes)

    def export(self):
        """Return a dictionary with all simulation parameters and objects.
        """
        js = {}
        js["parameters"] = write_parameters(self)
        js["sources"] = write_sources(self)
        js["monitors"] = write_monitors(self)
        js["materials"], js["structures"] = write_structures(self)

        return js

    def export_json(self, fjson):
        """Export the simulation dictionary to a JSON file.
        
        Parameters
        ----------
        fjson : str
            JSON file name.
        """

        self.fjson = fjson
        with open(fjson, 'w') as json_file:
            json.dump(self.export(), json_file, indent=4)

    @classmethod
    def import_json(cls, fjson):
        """Import a simulation from a JSON file.
        
        Parameters
        ----------
        fjson : str
            JSON file name.
        """
        
        with open(fjson, 'r') as json_file:
            js = json.load(json_file)

        sim = cls._read_simulation(js)
        sim.fjson = fjson

        return sim


import numpy as np
import logging

from .utils import listify
from .constants import int_, float_, EPSILON_0, C_0
from .dispersion import DispersionModel

class Medium(object):
    """
    Base class for a custom defined material.
    """

    def __init__(self, epsilon=1, sigma=0, name=None):
        """ Define a material through frequency-independent values of 
        the relative permittivity and the conductivity, or through a 
        dispersion function.
        
        Parameters
        ----------
        epsilon : float or DispersionModel, optional
            If numeric, real part of the relative permittivity. Default is
            vacuum. If a :class:`.DispersionModel` is provided, then the model 
            data is copied.
        sigma : float, optional
            (S/micron) Electric conductivity, s.t. 
            ``Im(eps(omega)) = sigma/omega``, where ``eps(omega)`` is the 
            complex permittivity at frequency omega. This conductivity is 
            added on top of any coming from the dispersion model, if used.
        """
        if isinstance(epsilon, DispersionModel):
            self.eps = epsilon._eps_inf
            self.sigma = 0
            self.poles = epsilon._poles
            self.dispmod = epsilon
        else:
            if 0 < epsilon < 1:
                warn_msg = (
                    "Permittivity smaller than one could result "
                    "in numerical instability. Use Courant stability factor "
                    "value lower than the smallest refractive index value."
                    )

            elif epsilon <= 0:
                warn_msg = (
                    "Permittivity smaller than zero can result in "
                    "numerical instability and should be included as a "
                    "dispersive model."
                    )

                if epsilon < -100:
                    warn_msg += \
                        "For large negative values consider using PEC instead."
                        
                logging.warning(warn_msg)

            self.dispmod = None
            self.eps = epsilon
            self.sigma = sigma
            self.poles = []
        
        # If set, this is a tuple (f_lower, f_upper) in Hz of the frequency
        # range of validity of this material model.
        self.frequency_range = None
        
        self.name = None if name is None else str(name)

    def epsilon(self, freq = None):
        """Evaluate the relative permittivity at a given frequency.
        
        Parameters
        ----------
        freq : None, optional
            (Hz) Frequency at which to evaluate the permittivity.
        
        Returns
        -------
        float
        """
        if self.dispmod is None:
            return self.eps * np.ones_like(freq)
        else:
            return self.dispmod.epsilon(freq)

    @classmethod
    def from_nk(cls, n, k, freq, name=None):
        """ Define a material through the real and imaginary part of the 
        refractive index at a given frequency.

        Parameters
        ----------
        n : float
            Real part of the refractive index.
        k : float
            Imaginary part of the refractive index.
        freq : float
            (Hz) Frequency at which ``(n, k)`` are evaluated.
        name : str, optional
            Custom name of the material.
        """
        eps_real = n**2 - k**2
        eps_imag = 2*n*k
        
        sigma = 2*np.pi*freq*eps_imag*EPSILON_0
        return cls(eps_real, sigma, name)

class PEC(object):
    """ Perfect electric conductor. All tangential electric fields vanish.
    """
    def __init__(self, name='PEC'):
        """ Construct.

        Parameters
        ----------
        name : str, optional
            Custom name of the material.
        """
        self.name=name

class PMC(object):
    """ Perfect magnetic conductor. All tangential magnetic fields vanish.
    """
    def __init__(self, name='PMC'):
        """ Construct.

        Parameters
        ----------
        name : str, optional
            Custom name of the material.
        """
        self.name=name




import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

def viz_source_spectrum(self, source, ax=None, Nfreqs=100, flims=None):
    """Plot the frequency dependence of a source.
    
    Parameters
    ----------
    source : Source
        Source object to plot.
    ax : Matplotlib axis object, optional
        If None, a new figure is created.
    Nfreqs : int
        Number of frequencies to interpolate the spectrum over.
    flims : List[float], optional
        Frequency limits (fmin, fmax) to plot within.
    
    Returns
    -------
    Matplotlib axis object
    """
    src_data = self._source_ids[id(source)]
    stime = source.source_time
    tdep = src_data.time_dep[:, 0]
    tmesh = src_data.tmesh

    if ax is None:
        fig, ax = plt.subplots(1)

    if flims is None:
        flims = (stime.frequency - 4*stime.fwidth,
                   stime.frequency + 4*stime.fwidth)

    # # Compute from FFT
    # Nt = tdep.size
    # dt = tmesh[1] - tmesh[0]
    # df = 1/Nt/dt
    # fmesh = np.arange(0, Nt)*df
    # spectrum = 1/np.sqrt(2*np.pi)*np.fft.ifft(tdep)/df

    # Compute directly from source time method
    fmesh = np.linspace(flims[0], flims[1], Nfreqs)
    spectrum = self.spectrum(source, fmesh)
    spectrum /= np.amax(np.abs(spectrum))
    pl = ax.plot(fmesh, np.abs(spectrum))

    ax.set_ylim([0, 1.1])
    ax.set_xlim(flims)
    ax.set_xlabel("Frequency [Hz]")
    ax.set_ylabel("Amplitude [a.u.]")
    ax.set_title("Spectrum")

    return ax

def viz_source_time(self, source, ax=None):
    """Plot the time dependence of a source.
    
    Parameters
    ----------
    source : Source
        Source object to plot.
    ax : Matplotlib axis object, optional
        If None, a new figure is created.

    Returns
    -------
    Matplotlib axis object
    """
    src_data = self._source_ids[id(source)]
    stime = source.source_time
    tdep = src_data.time_dep[:, 0]
    tmesh = src_data.tmesh
    tdep /= np.amax(np.abs(tdep))

    if ax is None:
        fig, ax = plt.subplots(1)

    pl = ax.plot(tmesh, tdep)
    ax.set_xlim([tmesh[0], tmesh[-1]])
    ax.set_ylim([-1.1, 1.1])
    ax.set_xlabel("Time [s]")
    ax.set_ylabel("Amplitude [a.u.]")
    ax.set_title("Time dependence");

    return ax

def viz_source(self, source, axs=None):
    """Plot both the time and frequency dependence of a source.
    
    Parameters
    ----------
    source : Source
        Source object to plot.
    axs : List[Matplotlib axis object], optional
        If None, a new figure is created.

    Returns
    -------
    Matplotlib axis object
    """
    
    if axs==None:
        fig, axs = plt.subplots(1, 2, figsize=(10, 3))

    self.viz_source_time(source, ax=axs[0])
    self.viz_source_spectrum(source, ax=axs[1])

    return axs


import numpy as np

from ..utils import inside_box, inside_box_inds, cs2span, listify, list2str
from ..utils import log_and_raise
from ..utils.check import check_3D_lists
from ..constants import int_, float_, complex_, fp_eps, xyz_list, xyz_dict, inf

class Source(object):

    def __init__(self, source_time, span, amplitude=1., name=None):
        """Base class sources. Available subclasses:

        - :class:`.VolumeSource`
        - :class:`.PointDipole`
        - :class:`.ModeSource`
        - :class:`.PlaneWave`
        """
        self.source_time = source_time
        self.span = span
        self.amplitude = amplitude
        self.name = None if name is None else str(name)
       
        # Nonzero components boolean: Ex, Ey, Ez, Hx, Hy, Hz. If 0, the solver 
        # automatically skips the component.
        self.components = np.zeros((6,), dtype=np.int8)

    def _inside(self, mesh):
        """ Get a mask equal to one if a point is inside the source region, 
        and zero if outside.
        
        Parameters
        ----------
        mesh : 3-tuple
            Defines the x, y, and z mesh. 
        """
        return inside_box(self.span, mesh, include_zero_size=True)

    def _inside_inds(self, mesh):
        """ Get indexes of the points inside the source region.
        
        Parameters
        ----------
        mesh : 3-tuple
            Defines the x, y, and z mesh. 
        
        Returns
        -------
        np.ndarray
            An array of shape (Np, 3), where Np is the total number of mesh 
            points in the source region.
        """
        inds = inside_box_inds(self.span, mesh, include_zero_size=True)
        indsm = np.meshgrid(np.arange(inds[0][0], inds[0][1]), 
                            np.arange(inds[1][0], inds[1][1]),
                            np.arange(inds[2][0], inds[2][1]),
                            indexing='ij')
        if indsm[0].size==0: 
            return np.zeros((0, 3), dtype=int_)

        return np.stack([inds.ravel() for inds in indsm], axis=1).astype(int_)

class VolumeSource(Source):
    """A source specified as an electric or a magnetic current of a fixed 
    component (one of ``Ex, Ey, Ez, Hx, Hy, Hz``) with a constant amplitude 
    inside a 3D volume region.
    """

    def __init__(self, source_time, center, size, component, 
                    amplitude=1., name=None):
        """
        Construct.
        
        Parameters
        ----------
        source_time : SourceTime
            Object describing the time dependence of the source.
        center : array_like
            (micron) 3D vector defining the center of the source region.
        size : array_like
            (micron) 3D vector defining the size of the source region.
        component : str
            One of ``{'Ex', 'Ey', 'Ez', 'Hx', 'Hy', 'Hz'}``, specifying the 
            direction and type of current component (e.g. ``Ez`` specifies 
            electric current source polarized along the z-axis).
        amplitude : float, optional
            Scaling amplitude of the source.
        name : None, optional
            Custom name of the source.
        """
        check_3D_lists(center=listify(center), size=listify(size))
        self.center = np.array(center)
        self.size = np.array(size)
        span = cs2span(center, size)
        super().__init__(source_time, span, amplitude, name)
        self.component = component
        cinds = ['ex', 'ey', 'ez', 'hx', 'hy', 'hz']
        try:
            comp_ind = cinds.index(component.lower())
            self.components[comp_ind] = 1
        except:
            log_and_raise(
                f"Unrecognized component {component}, must be one of {cinds}.",
                ValueError
            )

    def __repr__(self):
        rep = "Tidy3D VolumeSource: {\n"
        rep += "name         = %s\n"%self.name
        rep += "center       = %s\n" % list2str(self.center, "%1.4f")
        rep += "size         = %s\n" % list2str(self.size, "%1.4f")
        rep += "source_time  = " + repr(self.source_time) + "\n"
        rep += "component    = %s\n"%self.component
        rep += "amplitude    = %1.2e\n"%self.amplitude
        rep += "}\n"

        return rep

class PointDipole(VolumeSource):
    """A source specified as an electric or a magnetic current of a fixed 
    component (one of ``Ex, Ey, Ez, Hx, Hy, Hz``) at a given point.
    """

    def __init__(self, source_time, center, component, 
                    amplitude=1., name=None):
        """
        Construct.
        
        Parameters
        ----------
        source_time : SourceTime
            Object describing the time dependence of the source.
        center : array_like
            (micron) 3D vector defining the center of the source region.
        component : str
            One of ``{'Ex', 'Ey', 'Ez', 'Hx', 'Hy', 'Hz'}``, specifying the 
            direction and type of current component (e.g. ``Ez`` specifies 
            electric current source polarized along the z-axis).
        amplitude : float, optional
            Scaling amplitude of the source.
        name : None, optional
            Custom name of the source.
        """

        super().__init__(source_time,
                         center=center,
                         size=[0, 0, 0], 
                         component=component,
                         amplitude=amplitude,
                         name=name)

        def __repr__(self):
            rep = "Tidy3D PointDipole: {\n"
            rep += "name         = %s\n"%self.name
            rep += "center       = %s\n" % list2str(self.center, "%1.4f")
            rep += "source_time  = " + repr(self.source_time) + "\n"
            rep += "component    = %s\n"%self.component
            rep += "amplitude    = %1.2e\n"%self.amplitude
            rep += "}\n"

            return rep

class ModeSource(Source):
    """ Eigenmode source spanning a 2D cross section of a given size inside 
    a :class:`.Simulation`. The source needs to be added to a 
    :class:`.Simulation` object before the eigenmodes can be computed. They are 
    computed using the material parameters defined by the :class:`.Simulation` 
    in the 2D cross section, and the central frequency of the ModeSource time 
    dependence. Periodic boundary conditions are assumed in the 2D plane. 
    The modes are sorted by their effective index (propagation constant in the 
    normal direction), in decreasing order. Before the :class:`.Simulation` 
    can be run, the index of the eigenmode to be used by the ModeSource has to 
    be specified.
    """
    def __init__(self, source_time, center, size,
                    direction='forward', amplitude=1., name=None):
        """
        Construct.
        
        Parameters
        ----------
        source_time : SourceTime
            Object describing the time dependence of the source.
        center : array_like
            (micron) 3D vector defining the center of the 2D plane.
        size : array_like
            (micron) 3D vector defining the size of the 2D plane. Exactly one 
            of the values must be ``0``, defining the normal direction.
        direction : {'forward', 'backward'}
            Specifying propagation along the positive or negative direction of 
            the normal axis.
        amplitude : float, optional
            Scaling amplitude of the source.
        name : None, optional
            Custom name of the source.
        """

        check_3D_lists(center=listify(center), size=listify(size))
        # Get normal direction from size
        self.norm_ind = np.nonzero(size < fp_eps)[0]
        if self.norm_ind.size !=1:
            log_and_raise(
                "Exactly one element of 'size' must be zero.",
                ValueError
            )
        self.norm_ind = int(self.norm_ind)

        try:
            # Direction index is -/+ 1 for backward/forward
            dir_ind = ['backward', 'forward'].index(direction)
            self.dir_ind = int(2*(dir_ind - 0.5))
        except:
            log_and_raise(
                "'direction' must be one of 'forward', 'backward'.",
                ValueError
            )

        self.normal = xyz_list[self.norm_ind]
        self.direction = direction
        self.center = np.array(center)
        self.size = np.array(size)
        span = cs2span(center, size)
        super().__init__(source_time, span, amplitude, name)

        # All tangential components could be non-zero. 
        self.components = np.ones((6, ), dtype=np.int8)
        # No normal E component
        self.components[self.norm_ind] = 0
        # No normal H component
        self.components[3 + self.norm_ind] = 0

    def __repr__(self):
        rep = "Tidy3D ModeSource: {\n"
        rep += "name         = %s\n"%self.name
        rep += "normal       = %s\n"%self.normal
        rep += "direction    = %s\n"%self.direction
        rep += "center       = %s\n" % list2str(self.center, "%1.4f")
        rep += "size         = %s\n" % list2str(self.size, "%1.4f")
        rep += "source_time  = " + repr(self.source_time) + "\n"
        rep += "amplitude    = %1.2e\n"%self.amplitude

        return rep

class PlaneWave(ModeSource):
    """ Plane-wave source with a specified direction and polarization, spanning 
    the whole 2D simulation cross-section at a given position normal to a 
    given injection axis.
    """
    def __init__(self, source_time, injection_axis, position, polarization,
                    amplitude=1., name=None):
        """ Construct.

        Parameters
        ----------
        source_time : SourceTime
            Object describing the time dependence of the source.
        injection_axis : {'+x', '-x', '+y', '-y', '+z', '-z'}
            Defines the axis normal to the plane of the source, and the 
            direction of propagation of the plane wave.
        position : float
            (micron) Position of the plane along the nomral axis.
        polarization : {'x', 'y', 'z'}
            Electric field polarization of the plane wave. Cannot be the same 
            as the ``injection_axis``.
        amplitude : float, optional
            Scaling amplitude of the source.
        name : None, optional
            Custom name of the source.   
        """

        inj_axes = ['+x', '-x', '+y', '-y', '+z', '-z']
        if injection_axis not in inj_axes:
            log_and_raise(
                f"'injection_axis' must be one of {inj_axes}.",
                ValueError
            )
        norm_ind = xyz_dict[injection_axis[1]]
        direction = {'-': 'backward', '+': 'forward'}[injection_axis[0]]

        try:
            self.position = float(position)
        except:
            log_and_raise(
                "'position' is a single float defining the plane "
                "position along the 'normal' axis.",
                ValueError
            )
        try:
            self.epol_ind = xyz_dict[polarization]
        except:
            log_and_raise(
                "'polarization' must be one of 'x', 'y', or 'z'.",
                ValueError
            )

        if norm_ind==self.epol_ind:
            log_and_raise(
                "'polarization' orientation cannot be the same "
                "as 'injection_axis'.",
                ValueError
            )

        self.polarization = polarization
        self.injection_axis = injection_axis
    
        center = np.zeros((3,))
        center[norm_ind] = position
        size = inf*np.ones((3,))
        size[norm_ind] = 0
        super().__init__(source_time, center, size,
                            direction, amplitude=amplitude, name=None)

        # Re-initialize components
        self.components = np.zeros((6, ), dtype=np.int8)
        self.components[self.epol_ind] = 1
        # H-field polarization index (orthogonal to E and normal)
        self.hpol_ind = [0, 1, 2]
        self.hpol_ind.remove(self.epol_ind)
        self.hpol_ind.remove(self.norm_ind)
        self.hpol_ind = self.hpol_ind[0]
        # Magnetic field component (boolean)
        self.components[3 + self.hpol_ind] = 1

    def __repr__(self):
        rep = "Tidy3D PlaneWave:\n"
        rep += "name            = %s\n"%self.name
        rep += "injection_axis  = %s\n"%self.injection_axis
        rep += "position        = %1.2f\n"%self.position
        rep += "source_time     = " + repr(self.source_time) + "\n"
        rep += "polarization    = %s\n"%self.polarization
        rep += "amplitude       = %1.2e\n"%self.amplitude

        return rep

class PlaneSource(ModeSource):
    """ Plane-wave source with a specified direction and polarization, spanning 
    the whole simulation cross-section at a given position and axis normal.
    """
    def __init__(self, source_time, position, normal, direction='forward',
                    diff_order=[0, 0], pol_angle=0, polarization=None,
                    amplitude=1., name=None):
        """ Construct.

        Parameters
        ----------
        source_time : SourceTime
            A SourceTime object describing the time dependence of the source.
        normal : {'x', 'y', 'z'}
            Axis normal to the plane of the source.
        position : float
            (micron) Position of the plane along the nomral axis.
        direction : {'forward', 'backward'}
            Specifying propagation along the positive or negative direction of 
            the normal axis.
        diff_order : array_like
            Two integers specifying the diffraction order of the plane wave 
            Periodic (not Bloch) boundary conditions are currently always 
            assumed in the cross-sectional plane in which the source is placed.
        pol_angle : float
            (radian) Specifies the angle between the electric field 
            polarization of the plane wave and the plane defined by the normal 
            axis and the propagation direction. ``pol_angle=0`` (default) 
            specifies P polarization, while ``pol_angle=np.pi/2`` specifies S 
            polarization.
        polarization : array_like or None
            If provided, this overrides ``pol_angle``. A vector with three 
            elements specifically defining the E-field polarization vector. 
            If the polarization vector is not orthogonal to the k-vector, 
            only the Gram-Schmidt orthogonal part is used.
        name : None, optional
            Custom name of the source.   
        """

        try:
            self.norm_ind = xyz_dict[normal]
        except:
            log_and_raise(
                "'normal' must be one of 'x', 'y', or 'z'.",
                ValueError
            )

        try:
            self.position = float(position)
        except:
            log_and_raise(
                "'position' is a single float defining the plane "
                "position along the 'normal' axis.",
                ValueError
            )

        self.diff_order = listify(diff_order)
        self.pol_angle = pol_angle
        if polarization is None:
            self.polarization = np.array([0, 0, 0])
        else:
            check_3D_lists(polarization=polarization)
            self.polarization = np.array(polarization, dtype=float_)

        center = np.zeros((3,))
        center[self.norm_ind] = position
        size = inf*np.ones((3,))
        size[self.norm_ind] = 0
        super().__init__(source_time, center, size,
                    direction=direction, amplitude=amplitude, name=name)

    def __repr__(self):
        rep = "Tidy3D PlaneSource: {\n"
        rep += "name         = %s\n"%self.name
        rep += "position     = %1.2f\n"%self.position
        rep += "normal       = %s\n"%self.normal
        rep += "direction    = %s\n"%self.direction
        rep += "diff_order   = %s\n"%list2str(self.diff_order, "%d")
        if np.linalg.norm(self.polarization) < fp_eps:
            rep += "pol_angle    = %1.4f\n"%self.pol_angle
        else:
            rep += "polarization = %s\n"%list2str(self.polarization, "%1.2f")

        rep += "source_time  = " + repr(self.source_time) + "\n"
        rep += "amplitude    = %1.2e\n"%self.amplitude
        rep += "}\n"

        return rep


import numpy as np

from .structure import Box, Sphere, Cylinder, PolySlab, GdsSlab
from .source import VolumeSource, PointDipole, PlaneSource, PlaneWave, \
                    ModeSource, GaussianPulse
from .monitor import TimeMonitor, FreqMonitor, ModeMonitor
from .material import Medium, PEC, PMC
from .dispersion import DispersionModel
from .utils import listify, span2cs, cs2span, log_and_raise

def write_parameters(sim):
    """ Convert simulation parameters to a dict.
    """
    cent = sim.center
    # We only specify the size inside the pml when initializing simulation
    size = sim.size_in
    
    # Compute-weight prefactor per 1 billion (grid points * time steps)
    compute_factor = 0.5

    parameters = {
                "unit_length": "um",
                "unit_frequency": "THz",
                "unit_time": "ps",
                "x_cent": float(cent[0]),
                "y_cent": float(cent[1]),
                "z_cent": float(cent[2]),
                "x_span": float(size[0]),
                "y_span": float(size[1]),
                "z_span": float(size[2]),
                "mesh_step": sim.grid.mesh_step.tolist(),
                "symmetries": sim.symmetries, 
                "pml_layers": sim.pml_layers,
                "run_time": sim.run_time*1e12,
                "courant": sim.courant,
                "nodes": int(sim.Np),
                "time_steps": int(sim.Nt),
                "compute_weight": float(compute_factor*sim.Np*sim.Nt/1e9)
                }

    return parameters

def write_structures(sim):
    """ Convert all Structure objects to a list of text-defined objects, 
    and all the corresponding Material objects to a list of text-defined 
    materials. 
    """
    mat_list = []
    for material in sim.materials:
        if isinstance(material, Medium):
            poles = []
            for pole in material.poles:
                poles.append([pole[0].real, pole[0].imag,
                                pole[1].real, pole[1].imag])
            mat = {"type": "Medium",
                   "permittivity": float(material.eps),
                   "conductivity": float(material.sigma),
                   "poles": poles,
                   "name": material.name
                   }
            mat_list.append(mat)
        elif isinstance(material, PEC):
            mat_list.append({"type": "PEC"})
        elif isinstance(material, PMC):
            mat_list.append({"type": "PMC"})

    obj_list = []
    for istruct, structure in enumerate(sim.structures):
        obj = {
                "name": structure.name,
                "mat_index": sim.mat_inds[istruct]
                }
        if isinstance(structure, Box):
            cent, size = structure.center, structure.size
            obj.update({"type": "Box",
                        "x_cent": float(cent[0]),
                        "y_cent": float(cent[1]),
                        "z_cent": float(cent[2]),
                        "x_span": float(size[0]),
                        "y_span": float(size[1]),
                        "z_span": float(size[2])})
            obj_list.append(obj)
        elif isinstance(structure, Sphere):
            obj.update({"type": "Sphere",
                        "x_cent": float(structure.center[0]),
                        "y_cent": float(structure.center[1]),
                        "z_cent": float(structure.center[2]),
                        "radius": float(structure.radius)})
            obj_list.append(obj)
        elif isinstance(structure, Cylinder):
            obj.update({"type": "Cylinder",
                        "x_cent": float(structure.center[0]),
                        "y_cent": float(structure.center[1]),
                        "z_cent": float(structure.center[2]),
                        "axis": structure.axis,
                        "radius": float(structure.radius),
                        "height": float(structure.height)})
            obj_list.append(obj)
        elif isinstance(structure, PolySlab):
            obj.update({"type": "PolySlab",
                        "vertices": structure.vertices.tolist(),
                        "z_cent": float(structure.z_cent),
                        "z_size": float(structure.z_size)})
            obj_list.append(obj)

        # Split GdsSlab into PolySlab objects
        if isinstance(structure, GdsSlab):
            for (ip, verts) in enumerate(structure.gds_cell.get_polygons()):
                poly = obj.copy()
                if poly['name'] is None:
                    poly['name'] = 'struct%04d_poly%04d'%(istruct, ip)
                else:
                    poly['name'] += '_poly%04d'%ip
                poly.update({"type": "PolySlab",
                        "vertices": verts.tolist(),
                        "z_cent": float(structure.z_cent),
                        "z_size": float(structure.z_size)})
                obj_list.append(poly)
        
    return mat_list, obj_list

def write_sources(sim):
    src_list = []
    for source in sim.sources:
        src_time = {
            "type": "GaussianPulse",
            "frequency": source.source_time.frequency*1e-12, 
            "fwidth": source.source_time.fwidth*1e-12,
            "offset": source.source_time.offset,
            "phase": source.source_time.phase
            }
        if isinstance(source, VolumeSource):
            src = {
                "name": source.name,
                "type": "VolumeSource",
                "source_time": src_time,
                "center": source.center.tolist(),
                "size": source.size.tolist(),
                "component": source.component, 
                "amplitude": float(source.amplitude),
                }
        elif isinstance(source, PointDipole):
            src = {
                "name": source.name,
                "type": "PointDipole",
                "source_time": src_time,
                "center": source.center.tolist(),
                "size": source.size.tolist(),
                "component": source.component, 
                "amplitude": float(source.amplitude),
                }
        elif isinstance(source, PlaneWave):
            src = {
                "name": source.name,
                "type": "PlaneWave",
                "source_time": src_time,
                "injection_axis": source.injection_axis,
                "position": source.position,
                "polarization": source.polarization,
                "amplitude": float(source.amplitude)
                }
        elif isinstance(source, PlaneSource):
            src = {
                "name": source.name,
                "type": "PlaneSource",
                "source_time": src_time,
                "position": source.position,
                "normal": source.normal,
                "direction": source.direction,
                "diff_order": source.diff_order,
                "pol_angle": float(source.pol_angle),
                "polarization": source.polarization.tolist(),
                "amplitude": float(source.amplitude)
                }
        elif isinstance(source, ModeSource):
            src_data = sim._source_ids[id(source)]
            mode_ind = src_data.mode_ind
            if mode_ind is None:
                log_and_raise(
                    f"Mode index of source {src_data.name} not yet set, "
                    "use Simulation.set_mode().",
                    RuntimeError
                )
            src = {
                "name": source.name,
                "type": "ModeSource",
                "source_time": src_time,
                "center": source.center.tolist(),
                "size": source.size.tolist(),
                "direction": source.direction,
                "amplitude": float(source.amplitude),
                "mode_ind": int(mode_ind)
                }
        src_list.append(src)

    return src_list

def write_monitors(sim):
    mnt_list = []
    for monitor in sim.monitors:
        cent, size = span2cs(monitor.span)
        mnt = {"name": monitor.name,
                "x_cent": float(cent[0]),
                "y_cent": float(cent[1]),
                "z_cent": float(cent[2]),
                "x_span": float(size[0]),
                "y_span": float(size[1]),
                "z_span": float(size[2]),
                "field": monitor.field
                }
        if isinstance(monitor, TimeMonitor):
            mnt.update({
                "type": "TimeMonitor",
                "t_start": monitor.t_start,
                "t_stop": monitor.t_stop
                })
        elif isinstance(monitor, ModeMonitor):
            mnt.update({
                "type": "ModeMonitor",
                "frequency": [f*1e-12 for f in monitor.freqs]
                })
        elif isinstance(monitor, FreqMonitor):
            mnt.update({
                "type": "FrequencyMonitor",
                "frequency": [f*1e-12 for f in monitor.freqs]
                })
        mnt_list.append(mnt)

    return mnt_list

def read_structures(sim, js, scale_l=1):
    materials = []
    for mat in js['materials']:
        if mat['type'].lower()=='medium':
            poles = []
            for pole in mat['poles']:
                poles.append((pole[0]+pole[1]*1j, pole[2]+pole[3]*1j))
            if 0 == len(poles):
                materials.append(Medium(
                    epsilon = mat['permittivity'],
                    sigma = mat['conductivity']
                ))
            else:
                materials.append(Medium(
                    epsilon = DispersionModel(
                        eps_inf = mat['permittivity'],
                        poles = poles
                    ),
                    sigma = mat['conductivity']
                ))
        elif mat['type'].lower()=='pec':
            materials.append(PEC())
        elif mat['type'].lower()=='pmc':
            materials.append(PMC())

    for obj in js['structures']:
        if obj['type'].lower()=='box':
            cent = np.array([obj['x_cent'], obj['y_cent'],
                        obj['z_cent']])*scale_l
            size = np.array([obj['x_span'], obj['y_span'],
                        obj['z_span']])*scale_l
            sim.add(Box(center=cent, size=size,
                                material=materials[obj['mat_index']],
                                name=obj['name']))
        elif obj['type'].lower()=='sphere':
            cent = [obj['x_cent'], obj['y_cent'], obj['z_cent']]
            sim.add(Sphere(center=cent, radius=obj['radius'], 
                                material=materials[obj['mat_index']],
                                name=obj['name']))
        elif obj['type'].lower()=='cylinder':
            cent = [obj['x_cent'], obj['y_cent'], obj['z_cent']]
            sim.add(Cylinder(center=cent, axis=obj['axis'], 
                                radius=obj['radius'], height=obj['height'],
                                material=materials[obj['mat_index']],
                                name=obj['name']))
        elif obj['type'].lower()=='polyslab':
            sim.add(PolySlab(vertices=obj['vertices'],
                                z_cent=obj['z_cent'], z_size=obj['z_size'], 
                                material=materials[obj['mat_index']],
                                name=obj['name']))
        else:
            raise NotImplementedError("Unknown structure type " + obj['type'])

def read_sources(sim, js, scale_l=1, scale_f=1):
    sources = []
    js_params = js['parameters']
    for src in js['sources']:
        src_time = src['source_time']
        source_time = GaussianPulse(src_time['frequency']*scale_f, 
                                     src_time['fwidth']*scale_f,
                                     src_time['offset'],
                                     src_time['phase'])
        if src['type'].lower()=='planewave':
            psource = PlaneWave(source_time=source_time,
                            injection_axis=src['injection_axis'].lower(),
                            position=src['position'],
                            polarization=src['polarization'].lower(),
                            amplitude=src['amplitude'],
                            name=src['name'])
            sim.add(psource)
            sim.set_mode(psource, 0)
        elif src['type'].lower()=='planesource':
            psource = PlaneSource(source_time=source_time,
                            position=src['position'],
                            normal=src['normal'].lower(),
                            direction=src['direction'].lower(),
                            diff_order=src['diff_order'],
                            pol_angle=src['pol_angle'],
                            polarization=src['polarization'],
                            amplitude=src['amplitude'],
                            name=src['name'])
            sim.add(psource)
            sim.set_mode(psource, 0)
        elif src['type'].lower()=='pointdipole':
            sim.add(PointDipole(
                            source_time=source_time,
                            center=src['center'],
                            component=src['component'],
                            amplitude=src['amplitude'],
                            name=src['name']))
        elif src['type'].lower()=='volumesource':
            sim.add(VolumeSource(
                            source_time=source_time,
                            center=src['center'],
                            size=src['size'], 
                            component=src['component'],
                            amplitude=src['amplitude'],
                            name=src['name']))
        elif src['type'].lower()=='modesource':
            msource = ModeSource(
                            source_time=source_time,
                            center=src['center'],
                            size=src['size'],
                            direction=src['direction'].lower(),
                            amplitude=src['amplitude'],
                            name=src['name'])
            # Overwrite the mode index; this will then ensure that set_mode() 
            # is called when the source is added to a Simulation object.
            sim.add(msource)
            sim.set_mode(msource, src['mode_ind'])
        else:
            raise NotImplementedError("Unknown source type " + source['type'])

def read_monitors(sim, js, scale_l=1, scale_f=1):
    """ Read monitors from a json dictionary and load them to the Simulation 
    object sim.
    """

    for mnt in js['monitors']:
        cent = np.array([mnt['x_cent'], mnt['y_cent'],
                                mnt['z_cent']])*scale_l
        size = np.array([mnt['x_span'], mnt['y_span'],
                                mnt['z_span']])*scale_l
        if mnt['type'].lower()=='modemonitor':    
            sim.add(ModeMonitor(center=cent, size=size,
                        freqs=[f*scale_f for f in listify(mnt['frequency'])],
                        name=mnt['name']))
        elif mnt['type'].lower()=='frequencymonitor':    
            sim.add(FreqMonitor(center=cent, size=size,
                        freqs=[f*scale_f for f in listify(mnt['frequency'])],
                        field=mnt['field'],
                        name=mnt['name']))
        elif mnt['type'].lower()=='timemonitor':
            sim.add(TimeMonitor(center=cent, size=size,
                        field=mnt['field'],
                        t_start=mnt['t_start'],
                        t_stop=mnt['t_stop'],
                        name=mnt['name']))
        else:
            raise NotImplementedError("Unknown monitor type " + mnt['type'])

@classmethod
def _read_simulation(cls, js: dict):
    """ Initialize Simulation object based on JSON input dictionary ``js``.
    """

    # Unit scalings
    if js['parameters']['unit_length'].lower() == "mm":
        scale_l = 1e3
    elif js['parameters']['unit_length'].lower() == "um":
        scale_l = 1
    elif js['parameters']['unit_length'].lower() == "nm":
        scale_l = 1e-3

    if js['parameters']['unit_frequency'].lower() == "thz":
        scale_f = 1e12
    elif js['parameters']['unit_frequency'].lower() == "ghz":
        scale_f = 1e9
    elif js['parameters']['unit_frequency'].lower() == "mhz":
        scale_f = 1e6

    if js['parameters']['unit_time'].lower() == "fs":
        scale_t = 1e-15
    elif js['parameters']['unit_time'].lower() == "ps":
        scale_t = 1e-12
    elif js['parameters']['unit_time'].lower() == "ns":
        scale_t = 1e-9

    # Parse simulation parameters
    js_params = js['parameters']

    # Simulation span
    sim_cent = np.array([0., 0., 0.])
    if "x_cent" in js_params.keys():
        sim_cent[0] = js_params["x_cent"]*scale_l
    if "y_cent" in js_params.keys():
        sim_cent[1] = js_params["y_cent"]*scale_l
    if "z_cent" in js_params.keys():
        sim_cent[2] = js_params["z_cent"]*scale_l

    sim_size = scale_l*np.array([js_params["x_span"], js_params["y_span"],
                                js_params["z_span"]])

    # Pml size
    pml_layers = np.array([0, 0, 0])
    if "pml_layers" in js_params.keys():
        pml_layers = js_params["pml_layers"]
    else:
        # If pml_layers not set directly, we assume BCs are defined
        for (i, si) in enumerate(['x', 'y', 'z']):
            if js_params[si+'_bc'].lower()=='pml':
                # set some default PML thickness
                pml_layers[i] = 12


    sim_params = {
            'center': sim_cent,
            'size': sim_size, 
            'mesh_step': js_params['mesh_step'],
            'run_time': js_params['run_time']*scale_t,
            'pml_layers': pml_layers
            }

    for key in ['symmetries', 'courant']:
        try:
            sim_params[key] = js_params[key]
        except:
            pass

    sim = cls(**sim_params)

    if "structures" in js.keys():
        read_structures(sim, js, scale_l=scale_l)
    if "sources" in js.keys():
        read_sources(sim, js, scale_l, scale_f)
    if "monitors" in js.keys():
        read_monitors(sim, js, scale_l, scale_f)
        
    return sim


import numpy as np

from .geom import intersect_box
from .log import log_and_raise

# Constants defining maximum size, etc.
MAX_TIME_STEPS = 1e8
MAX_GRID_CELLS = 4e9
MAX_CELLS_STEPS = 1e15 # max product of grid cells and time steps
MAX_MONITOR_DATA = 10  # in Gb

def check_3D_lists(**kwargs):
    """ Verify that input arguments are lists with three elements """
    for key, val in kwargs.items():
        try:
            if not isinstance(val, list) and not isinstance(val, tuple):
                raise ValueError
            if len(val) != 3:
                raise ValueError
            for v in val:
                if type(v) in [list, tuple, np.ndarray]:
                    raise ValueError
        except:
            log_and_raise (
                f"'{key}' must be array-like with three elements.",
                ValueError
            )

def _check_size(self):
    """ Check the size of a simulation vs. pre-defined maximum allowed values. 
    """

    if self.Nt > MAX_TIME_STEPS:
        log_and_raise(
            f"Time steps {self.Nt:.2e} exceed current limit "
            f"{MAX_TIME_STEPS:.2e}, reduce 'run_time' or increase the "
            "spatial mesh step.",
            RuntimeError
        )

    if self.Np > MAX_GRID_CELLS:
        log_and_raise(
            f"Total number of grid points {self.Np:.2e} exceeds "
            f"current limit {MAX_GRID_CELLS:.2e}, increase the mesh step "
            "or decrease the size of the simulation domain.",
            RuntimeError
        )

    if self.Np * self.Nt > MAX_CELLS_STEPS:
        log_and_raise(
            f"Product of grid points and time steps {self.Np*self.Nt:.2e} "
            f"exceeds current limit {MAX_CELLS_STEPS:.2e}. Increase the "
            "mesh step and/or decrease the 'run_time' of the simulation.",
            RuntimeError
        )

def _check_monitor_size(self, monitor):
    """ Check if expected monitor data is too big.
    """

    from ..monitor import TimeMonitor, FreqMonitor

    # Compute how many grid points there are inside the monitor
    mnt_data = self._monitor_ids[id(monitor)]
    span_in = intersect_box(self.span_sym, monitor.span)
    size_in = span_in[:, 1] - span_in[:, 0]
    if np.any(size_in < 0):
        Np = 0
    else:
        Np = np.prod([int(s)/self.grid.mesh_step[d] + 1 
                for (d, s) in enumerate(size_in)])

    if isinstance(monitor, TimeMonitor):
        mnt_data._set_tmesh(self.grid.tmesh)
        # 4 bytes x N points x N time steps x 3 components x N fields
        memGB = 4*Np*mnt_data.Nt*3*len(monitor.field)/1e9
        if memGB > MAX_MONITOR_DATA:
            log_and_raise(
                f"Estimated time monitor size {memGB:.2f}GB exceeds "
                f"current limit of {MAX_MONITOR_DATA:.2f}GB per monitor. "
                "Decrease monitor size or the time interval using "
                "'t_start' and 't_stop'.",
                RuntimeError
            )
                
    elif isinstance(monitor, FreqMonitor):
        # 8 bytes x N points x N freqs x 3 components x N fields
        memGB = 8*Np*len(mnt_data.freqs)*3*len(monitor.field)/1e9
        if memGB > MAX_MONITOR_DATA:
            log_and_raise(
                f"Estimated frequency monitor size {memGB:.2f}GB exceeds "
                f"current limit of {MAX_MONITOR_DATA:.2f}GB per monitor. " 
                "Decrease monitor size or the number of frequencies.",
                RuntimeError
            )

    return memGB